"""A multi-objective process with different search and solution spaces."""

from math import isfinite
from typing import Callable, Final

import numpy as np
from numpy import copyto

from moptipy.api._mo_process_no_ss import _MOProcessNoSS
from moptipy.api._process_base import _TIME_IN_NS
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.logging import (
    PREFIX_SECTION_ARCHIVE,
    SCOPE_ENCODING,
    SCOPE_SEARCH_SPACE,
    SECTION_RESULT_X,
    SUFFIX_SECTION_ARCHIVE_X,
    SUFFIX_SECTION_ARCHIVE_Y,
)
from moptipy.api.mo_archive import MOArchivePruner, MORecord
from moptipy.api.mo_problem import MOProblem
from moptipy.api.space import Space, check_space
from moptipy.utils.logger import KeyValueLogSection, Logger
from moptipy.utils.path import Path
from moptipy.utils.types import type_error


class _MOProcessSS(_MOProcessNoSS):
    """A class implementing a process with search and solution space."""

    def __init__(self,
                 solution_space: Space,
                 objective: MOProblem,
                 algorithm: Algorithm,
                 pruner: MOArchivePruner,
                 archive_max_size: int,
                 archive_prune_limit: int,
                 log_file: Path | None = None,
                 search_space: Space = None,
                 encoding: Encoding = None,
                 rand_seed: int | None = None,
                 max_fes: int | None = None,
                 max_time_millis: int | None = None,
                 goal_f: int | float | None = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param pruner: the archive pruner
        :param archive_max_size: the maximum archive size after pruning
        :param archive_prune_limit: the archive size above which pruning will
            be performed
        :param search_space: the search space.
        :param encoding: the encoding
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the
            process is terminated
        """
        super().__init__(solution_space=solution_space,
                         objective=objective,
                         algorithm=algorithm,
                         pruner=pruner,
                         archive_max_size=archive_max_size,
                         archive_prune_limit=archive_prune_limit,
                         log_file=log_file,
                         rand_seed=rand_seed,
                         max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)

        #: The search space.
        self._search_space: Final[Space] = check_space(search_space)
        #: The encoding.
        self._encoding: Final[Encoding] = check_encoding(encoding)
        #: the internal encoder
        self._g: Final[Callable] = encoding.decode
        #: The holder for the currently de-coded solution.
        self._current_y = solution_space.create()
        #: The current best point in the search space.
        self._current_best_x: Final = search_space.create()
        # wrappers
        self.create = search_space.create  # type: ignore
        self.copy = search_space.copy  # type: ignore
        self.to_str = search_space.to_str  # type: ignore
        self.is_equal = search_space.is_equal  # type: ignore
        self.from_str = search_space.from_str  # type: ignore
        self.n_points = search_space.n_points  # type: ignore
        self.validate = search_space.validate  # type: ignore
        self._create_y = solution_space.create  # the y creator

    def f_evaluate(self, x, fs: np.ndarray) -> float | int:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and the "
                                 "algorithm knows it.")
            return self._current_best_f

        current_y: Final = self._current_y
        self._g(x, current_y)
        result: Final[int | float] = self._f_evaluate(current_y, fs)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        improved: bool = False
        if result < self._current_best_f:
            improved = True
            self._current_best_f = result
            copyto(self._current_best_fs, fs)
            self.copy(self._current_best_x, x)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            do_term = do_term or (result <= self._end_f)

        if self.check_in(x, fs, True) or improved:
            self._last_improvement_fe = current_fes
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn

        if do_term:
            self.terminate()

        return result

    def get_copy_of_best_x(self, x) -> None:
        if self._current_fes > 0:
            return self.copy(x, self._current_best_x)
        raise ValueError("No current best x available.")

    def get_copy_of_best_y(self, y) -> None:
        if self._current_fes > 0:
            return self._copy_y(y, self._current_best_y)
        raise ValueError("No current best y available.")

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_SEARCH_SPACE) as sc:
            self._search_space.log_parameters_to(sc)
        with logger.scope(SCOPE_ENCODING) as sc:
            self._encoding.log_parameters_to(sc)

    def _write_result(self, logger: Logger) -> None:
        with logger.text(SECTION_RESULT_X) as txt:
            txt.write(self._search_space.to_str(self._current_best_x))
        super()._write_result(logger)

    def _validate_x(self) -> None:
        """Validate x, if it exists."""
        self._search_space.validate(self._current_best_x)

    def _log_and_check_archive_entry(self, index: int, rec: MORecord,
                                     logger: Logger) -> int | float:
        """
        Write an archive entry.

        :param index: the index of the entry
        :param rec: the record to verify
        :param logger: the logger
        :returns: the objective value
        """
        self.validate(rec.x)
        self.f_validate(rec.fs)
        tfs: Final[np.ndarray] = self._fs_temp

        current_y: Final = self._current_y
        self._g(rec.x, current_y)
        self._solution_space.validate(current_y)
        f: Final[int | float] = self._f_evaluate(current_y, tfs)

        if not np.array_equal(tfs, rec.fs):
            raise ValueError(
                f"expected {rec.fs} but got {tfs} when re-evaluating {rec}")
        if not isinstance(f, int | float):
            raise type_error(f, "scalarized objective value", (int, float))
        if not isfinite(f):
            raise ValueError(f"scalarized objective value {f} is not finite")

        with logger.text(f"{PREFIX_SECTION_ARCHIVE}{index}"
                         f"{SUFFIX_SECTION_ARCHIVE_X}") as lg:
            lg.write(self.to_str(rec.x))

        with logger.text(f"{PREFIX_SECTION_ARCHIVE}{index}"
                         f"{SUFFIX_SECTION_ARCHIVE_Y}") as lg:
            lg.write(self._solution_space.to_str(current_y))

        return f

    def __str__(self) -> str:
        return "MOProcessWithSearchSpace"



"""Minimize the sum of the work times of all machines in `Gantt` charts."""
from typing import Final

import numba  # type: ignore

from moptipy.api.objective import Objective
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.instance import Instance
from moptipy.utils.types import type_error


@numba.njit(nogil=True, cache=True)
def worktime(x: Gantt) -> int:
    """
    Get the work time of all machines in a given `Gantt` chart.

    We assume that a machine is turned on when its first job/operation begins
    and is turned off when the last job/operation on it ends.
    During all the time inbetween, the machine is on and a worker needs to be
    present.
    The work time therefore corresponds to the sum of the time units when the
    machines can be turned off minus the time units at which they are turned
    on.

    :param x: the Gantt chart.
    :return: the end times minus the start times
    """
    return int(x[:, -1, 2].sum()) - int(x[:, 0, 2].sum())


class Worktime(Objective):
    """Compute the work time of a `Gantt` chart (for minimization)."""

    def __init__(self, instance: Instance) -> None:  # +book
        """
        Initialize the worktime objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        self.__instance: Final[Instance] = instance
        #: The fast call forwarding to the worktime function. # +book
        self.evaluate = worktime  # type: ignore # +book

    def lower_bound(self) -> int:
        """
        Get the lower bound of the worktime.

        :return: the lower bound
        """
        return self.__instance.makespan_lower_bound \
            + self.__instance.machines - 1

    def is_always_integer(self) -> bool:
        """
        Return `True` because :func:`worktime` always returns `int` values.

        :retval True: always
        """
        return True

    def upper_bound(self) -> int:
        """
        Get the upper bound of the worktime.

        :return: the sum of all job execution times.
        """
        return self.__instance.makespan_upper_bound * self.__instance.machines

    def __str__(self) -> str:
        """
        Get the name of the worktime objective function.

        :return: `worktime`
        :retval "worktime": always
        """
        return "worktime"



"""
Search operators for permutations with repetitions.

These operators can be used in conjunction with the space
:class:`~moptipy.spaces.permutations.Permutations`.

- Module :mod:`~moptipy.operators.permutations.op0_shuffle` provides a nullary
  operator sampling an entirely random permutation.
- Module :mod:`~moptipy.operators.permutations.op1_swap2` provides a unary
  operator which swaps exactly two different elements of a permutation.
- Module :mod:`~moptipy.operators.permutations.op1_swapn` provides a unary
  operator which tries to find a cyclic move whose length would be
  approximately binomially distributed, i.e., which may swap two elements with
  probability 0.5, three elements with probability 0.25, 4 elements with
  probability 0.125, and so on. It does so by performing a sequence of swaps
  but does not check whether one swap undoes a previous one.
- Module :mod:`~moptipy.operators.permutations.op1_swap_exactly_n` offers a
  unary operator with a step size. A step size of 0 means that it will swap
  two elements, a step size of 1 means that it tries to perform the
  largest-possible modification. If it is applied to permutations where each
  element occurs once, this means that all elements change their position. If
  applied to permutations with repetitions, things are more complex and some
  moves may be impossible, in which case a best effort is attempted.
- Module :mod:`~moptipy.operators.permutations.op2_gap` offers an operator
  that tries to build a new permutation by appending not-yet-appended elements
  from both input permutations, alternating between them randomly.
- Module :mod:`~moptipy.operators.permutations.op2_ox2` implements a slightly
  modified version of the order-based crossover operator for permutations.
"""



"""
A temperature schedule as needed by Simulated Annealing.

The Simulated Annealing algorithm implemented in
:mod:`~moptipy.algorithms.so.simulated_annealing` performs a local search that
always accepts a non-worsening move, i.e., a solution which is not worse than
the currently maintained one. However, it will also *sometimes* accept one
that is worse. The probability of doing so depends on how much worse that
solution is and on the current *temperature* of the algorithm. The higher the
temperature, the higher the acceptance probability. The temperature changes
over time according to the
:class:`~moptipy.algorithms.modules.temperature_schedule.TemperatureSchedule`.

The temperature schedule receives an iteration index `tau` as input and
returns the current temperature via :meth:`~moptipy.algorithms.modules.\
temperature_schedule.TemperatureSchedule.temperature`. Notice that `tau` is
zero-based for simplicity reason, meanings that the first objective function
evaluation is at index `0`.
"""

from math import e, isfinite, log
from typing import Final

from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error


# start schedule
class TemperatureSchedule(Component):
    """The base class for temperature schedules."""

    def __init__(self, t0: float) -> None:
        # end schedule
        """
        Initialize the temperature schedule.

        :param t0: the starting temperature, must be > 0
        """
        super().__init__()
        if not isinstance(t0, float):
            raise type_error(t0, "t0", float)
        if (not isfinite(t0)) or (t0 <= 0.0):
            raise ValueError(f"t0 must be >0, cannot be {t0}.")
# start schedule
        #: the starting temperature
        self.t0: Final[float] = t0

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature
        """
# end schedule

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this temperature schedule as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         TemperatureSchedule(0.1).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: TemperatureSchedule'
        >>> text[3]
        'T0: 0.1'
        >>> len(text)
        6
        """
        super().log_parameters_to(logger)
        logger.key_value("T0", self.t0)


# start exponential
class ExponentialSchedule(TemperatureSchedule):
    """
    The exponential temperature schedule.

    The current temperature is computed as `t0 * (1 - epsilon) ** tau`.

    >>> ex = ExponentialSchedule(10.0, 0.05)
    >>> print(f"{ex.t0} - {ex.epsilon}")
    10.0 - 0.05
    >>> ex.temperature(0)
    10.0
    >>> ex.temperature(1)
    9.5
    >>> ex.temperature(2)
    9.025
    >>> ex.temperature(1_000_000_000_000_000_000)
    0.0
    """

    def __init__(self, t0: float, epsilon: float) -> None:
        """
        Initialize the exponential temperature schedule.

        :param t0: the starting temperature, must be > 0
        :param epsilon: the epsilon parameter of the schedule, in (0, 1)
        """
        super().__init__(t0)
# end exponential
        if not isinstance(epsilon, float):
            raise type_error(epsilon, "epsilon", float)
        if (not isfinite(epsilon)) or (not (0.0 < epsilon < 1.0)):
            raise ValueError(
                f"epsilon cannot be {epsilon}, must be in (0,1).")
# start exponential
        #: the epsilon parameter of the exponential schedule
        self.epsilon: Final[float] = epsilon
        #: the value used as basis for the exponent
        self.__one_minus_epsilon: Final[float] = 1.0 - epsilon
# end exponential
        if not (0.0 < self.__one_minus_epsilon < 1.0):
            raise ValueError(
                f"epsilon cannot be {epsilon}, because 1-epsilon must be in "
                f"(0, 1) but is {self.__one_minus_epsilon}.")
# start exponential

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature

        >>> s = ExponentialSchedule(100.0, 0.5)
        >>> s.temperature(0)
        100.0
        >>> s.temperature(1)
        50.0
        >>> s.temperature(10)
        0.09765625
        """
        return self.t0 * (self.__one_minus_epsilon ** tau)
# end exponential

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the exponential temperature schedule.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         ExponentialSchedule(0.2, 0.6).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: exp0d2_0d6'
        >>> text[3]
        'T0: 0.2'
        >>> text[5]
        'e: 0.6'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value("e", self.epsilon)

    def __str__(self) -> str:
        """
        Get the string representation of the exponential temperature schedule.

        :returns: the name of this schedule

        >>> ExponentialSchedule(100.5, 0.3)
        exp100d5_0d3
        """
        return f"exp{num_to_str_for_name(self.t0)}_" \
               f"{num_to_str_for_name(self.epsilon)}"


# start logarithmic
class LogarithmicSchedule(TemperatureSchedule):
    """
    The logarithmic temperature schedule.

    The temperature is computed as `t0 / log(e + (tau * epsilon))`.

    >>> lg = LogarithmicSchedule(10.0, 0.1)
    >>> print(f"{lg.t0} - {lg.epsilon}")
    10.0 - 0.1
    >>> lg.temperature(0)
    10.0
    >>> lg.temperature(1)
    9.651322627630812
    >>> lg.temperature(1_000_000_000_000_000_000_000_000_000_000_000_000_000)
    0.11428802155348732
    """

    def __init__(self, t0: float, epsilon: float) -> None:
        """
        Initialize the logarithmic temperature schedule.

        :param t0: the starting temperature, must be > 0
        :param epsilon: the epsilon parameter of the schedule, is > 0
        """
        super().__init__(t0)
# end logarithmic
        if not isinstance(epsilon, float):
            raise type_error(epsilon, "epsilon", float)
        if (not isfinite(epsilon)) or (epsilon <= 0.0):
            raise ValueError(
                f"epsilon cannot be {epsilon}, must be > 0.")
# start logarithmic
        #: the epsilon parameter of the logarithmic schedule
        self.epsilon: Final[float] = epsilon

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature

        >>> s = LogarithmicSchedule(100.0, 0.5)
        >>> s.temperature(0)
        100.0
        >>> s.temperature(1)
        85.55435113150568
        >>> s.temperature(10)
        48.93345190925178
        """
        return self.t0 / log(e + (tau * self.epsilon))
# end logarithmic

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the logarithmic temperature schedule.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         LogarithmicSchedule(0.2, 0.6).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: ln0d2_0d6'
        >>> text[3]
        'T0: 0.2'
        >>> text[5]
        'e: 0.6'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value("e", self.epsilon)

    def __str__(self) -> str:
        """
        Get the string representation of the logarithmic temperature schedule.

        :returns: the name of this schedule

        >>> LogarithmicSchedule(100.5, 0.3)
        ln100d5_0d3
        """
        return f"ln{num_to_str_for_name(self.t0)}_" \
               f"{num_to_str_for_name(self.epsilon)}"



"""
Classes for writing structured log files.

A :class:`~Logger` offers functionality to write structured, text-based log
files that can hold a variety of information. It is implemented in two
flavors, :class:`~FileLogger`, which writes data to a file, and
:class:`~InMemoryLogger`, which writes data to a buffer in memory (which is
mainly useful for testing).

A :class:`~Logger` can produce output in three formats:

- :meth:`~Logger.csv` creates a section of semicolon-separated-values data
  (:class:`~CsvLogSection`), which we call `csv` because it structured
  basically exactly as the well-known comma-separated-values data, just
  using semicolons.
- :meth:`~Logger.key_values` creates a key-values section
  (:class:`~KeyValueLogSection`) in YAML format. The specialty of this
  section is that it permits hierarchically structuring of data by spawning
  out sub-sections that are signified by a key prefix via
  :meth:`~KeyValueLogSection.scope`.
- :meth:`~Logger.text` creates a raw text section (:class:`~TextLogSection`),
  into which raw text can be written.

The beginning and ending of section named `XXX` are `BEGIN_XXX` and `END_XXX`.

This class is used by the :class:`~moptipy.api.execution.Execution` and
experiment-running facility (:func:`~moptipy.api.experiment.run_experiment`)
to produce log files complying with
https://thomasweise.github.io/moptipy/#log-files.
Such log files can be parsed via :mod:`~moptipy.evaluation.log_parser`.
"""

from contextlib import AbstractContextManager
from io import StringIO, TextIOBase
from math import isfinite
from os.path import realpath
from re import sub
from typing import Callable, Final, Iterable, cast

from moptipy.utils.cache import is_new
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    PART_SEPARATOR,
    bool_to_str,
    float_to_str,
    sanitize_name,
)
from moptipy.utils.types import type_error

#: the separator used in CSV files to separate columns
CSV_SEPARATOR: Final[str] = ";"
#: the character indicating the begin of a comment
COMMENT_CHAR: Final[str] = "#"
#: the character separating a scope prefix in a key-value section
SCOPE_SEPARATOR: Final[str] = "."
#: the indicator of the start of a log section
SECTION_START: Final[str] = "BEGIN_"
#: the indicator of the end of a log section
SECTION_END: Final[str] = "END_"
#: the replacement for special characters
SPECIAL_CHAR_REPLACEMENT: Final[str] = PART_SEPARATOR
#: the YAML-conform separator between a key and a value
KEY_VALUE_SEPARATOR: Final[str] = ": "
#: the hexadecimal version of a value
KEY_HEX_VALUE: Final[str] = "(hex)"


class Logger(AbstractContextManager):
    """
    An abstract base class for logging data in a structured way.

    There are two implementations of this, :class:`InMemoryLogger`, which logs
    data in memory and is mainly there for testing and debugging, an
    :class:`FileLogger` which logs to a text file and is to be used in
    experiments with `moptipy`.
    """

    def __init__(self, stream: TextIOBase, name: str) -> None:
        """
        Create a new logger.

        :param stream: the stream to which we will log, will be closed when
            the logger is closed
        :param name: the name of the logger
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if stream is None:
            raise ValueError("stream must be valid stream but is None.")
        if not isinstance(stream, TextIOBase):
            raise type_error(stream, "stream", TextIOBase)

        #: The internal stream
        self._stream: TextIOBase = stream
        self.__section: str | None = None
        self.__starts_new_line: bool = True
        self.__log_name: str = name
        self.__sections: Callable = is_new()
        self.__closer: str | None = None

    def __enter__(self):
        """
        Enter the logger in a `with` statement.

        :return: `self`
        """
        return self

    def _error(self, message: str) -> None:
        """
        Raise a :class:`ValueError` with context information.

        :param message: the message elements to merge
        :raises ValueError: an error with the message and some context
        information
        """
        raise ValueError(f"{message} in logger {self.__log_name!r}."
                         if self.__section is None else
                         f"{message} in section {self.__section!r} "
                         f"of logger {self.__log_name!r}.")

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """
        Close the logger after leaving the `with` statement.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        """
        if self.__section is not None:
            self._error("Cannot close logger, because section still open")
        if self._stream is not None:
            if not self.__starts_new_line:
                self._stream.write("\n")
            self._stream.close()
            self._stream = None

    def _open_section(self, title: str) -> None:
        """
        Open a new section.

        :param title: the section title
        """
        if self._stream is None:
            self._error(f"Cannot open section {title!r} "
                        "because logger already closed")

        if self.__section is not None:
            self._error(f"Cannot open section {title!r} because "
                        "another one is open")

        new_title = title.strip().upper()
        if new_title != title:
            self._error(f"Cannot open section {title!r} because "
                        "title is invalid")

        if not self.__sections(title):
            self._error(f"Section {title!r} already done")

        self._stream.write(f"{SECTION_START}{title}\n")
        self.__closer = f"{SECTION_END}{title}\n"
        self.__starts_new_line = True
        self.__section = title

    def _close_section(self, title: str) -> None:
        """
        Close a section.

        :param title: the section title
        """
        if (self.__section is None) or (self.__section != title):
            self._error(f"Cannot open section {title!r} since it is not open")
        printer = self.__closer
        if not self.__starts_new_line:
            printer = "\n" + printer

        self._stream.write(printer)
        self.__closer = None
        self.__starts_new_line = True
        self.__section = None

    def _comment(self, comment: str) -> None:
        """
        Write a comment.

        :param comment: the comment
        """
        if self.__section is None:
            self._error("Cannot write if not inside section")
        if len(comment) <= 0:
            return
        comment = sub(r"\s+", " ", comment.strip())
        self._stream.write(f"{COMMENT_CHAR} {comment}\n")
        self.__starts_new_line = True

    def _write(self, text: str) -> None:
        """
        Write a string.

        :param text: the text to write
        """
        if self.__section is None:
            self._error("Cannot write if not inside section")

        if len(text) <= 0:
            return

        if self.__closer in text:
            self._error(f"String {self.__closer!r} "
                        "must not be contained in output")

        text = text.replace("#", "")  # omit all # characters
        self._stream.write(text)
        self.__starts_new_line = text.endswith("\n")

    def key_values(self, title: str) -> "KeyValueLogSection":
        r"""
        Create a log section for key-value pairs.

        The contents of such a section will be valid YAML mappings, i.w.,
        conform to
        https://yaml.org/spec/1.2/spec.html#mapping.
        This means they can be parsed with a YAML parser (after removing the
        section start and end marker, of course).

        :param title: the title of the new section
        :return: the new logger

        >>> from moptipy.utils.temp import TempFile
        >>> with TempFile.create() as t:
        ...     with FileLogger(str(t)) as l:
        ...         with l.key_values("B") as kv:
        ...             kv.key_value("a", "b")
        ...             with kv.scope("c") as kvc:
        ...                 kvc.key_value("d", 12)
        ...                 kvc.key_value("e", True)
        ...             kv.key_value("f", 3)
        ...     text = open(str(t), "r").read().splitlines()
        >>> print(text)
        ['BEGIN_B', 'a: b', 'c.d: 12', 'c.e: T', 'f: 3', 'END_B']
        >>> import yaml
        >>> dic = yaml.safe_load("\n".join(text[1:5]))
        >>> print(list(dic.keys()))
        ['a', 'c.d', 'c.e', 'f']
        """
        return KeyValueLogSection(title=title, logger=self, prefix="",
                                  done=None)

    def csv(self, title: str, header: list[str]) -> "CsvLogSection":
        """
        Create a log section for CSV data with `;` as column separator.

        The first line will be the headline with the column names.

        :param title: the title of the new section
        :param header: the list of column titles
        :return: the new logger

        >>> from moptipy.utils.logger import FileLogger
        >>> from moptipy.utils.temp import TempFile
        >>> with TempFile.create() as t:
        ...     with FileLogger(str(t)) as l:
        ...         with l.csv("A", ["x", "y"]) as csv:
        ...             csv.row([1,2])
        ...             csv.row([3,4])
        ...             csv.row([4, 12])
        ...     text = open(str(t), "r").read().splitlines()
        ...     print(text)
        ['BEGIN_A', 'x;y', '1;2', '3;4', '4;12', 'END_A']
        >>> import csv
        >>> for r in csv.reader(text[1:5], delimiter=";"):
        ...     print(r)
        ['x', 'y']
        ['1', '2']
        ['3', '4']
        ['4', '12']
        """
        return CsvLogSection(title=title, logger=self, header=header)

    def text(self, title: str) -> "TextLogSection":
        r"""
        Create a log section for unstructured text.

        :param title: the title of the new section
        :return: the new logger

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.text("C") as tx:
        ...         tx.write("aaaaaa")
        ...         tx.write("bbbbb")
        ...         tx.write("\n")
        ...         tx.write("ccccc")
        ...     print(l.get_log())
        ['BEGIN_C', 'aaaaaabbbbb', 'ccccc', 'END_C']
        """
        return TextLogSection(title=title, logger=self)


class FileLogger(Logger):
    """A logger logging to a file."""

    def __init__(self, path: str) -> None:
        """
        Initialize the logger.

        :param path: the path to the file to open
        """
        if not isinstance(path, str):
            raise type_error(path, "path", str)
        name = path
        path = realpath(path)
        super().__init__(stream=Path.path(path).open_for_write(),
                         name=name)


class InMemoryLogger(Logger):
    """A logger logging to a string in memory."""

    def __init__(self) -> None:
        """Initialize the logger."""
        super().__init__(stream=StringIO(),
                         name="in-memory-logger")

    def get_log(self) -> list[str]:
        """
        Obtain all the lines logged to this logger.

        :return: a list of strings with the logged lines
        """
        return cast(StringIO, self._stream).getvalue().splitlines()


class LogSection(AbstractContextManager):
    """An internal base class for logger sections."""

    def __init__(self, title: str | None, logger: Logger) -> None:
        """
        Perform internal construction. Do not call directly.

        :param title: the section title
        :param logger: the logger
        """
        self._logger: Logger = logger
        self._title: str | None = title
        if title is not None:
            # noinspection PyProtectedMember
            logger._open_section(title)

    def __enter__(self):
        """
        Enter the context: needed for the `with` statement.

        :return: `self`
        """
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """
        Exit the `with` statement.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :return: ignored
        """
        if self._title is not None:
            # noinspection PyProtectedMember
            self._logger._close_section(self._title)
            self._title = None
        self._logger = None

    def comment(self, comment: str) -> None:
        """
        Write a comment line.

        A comment starts with `#` and is followed by text.

        :param comment: the comment to write

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.text("A") as tx:
        ...         tx.write("aaaaaa")
        ...         tx.comment("hello")
        ...     print(l.get_log())
        ['BEGIN_A', 'aaaaaa# hello', 'END_A']
        """
        # noinspection PyProtectedMember
        self._logger._comment(comment)


class CsvLogSection(LogSection):
    """
    A logger that is designed to output CSV data.

    The coma-separated-values log is actually a semicolon-separated-values
    log. This form of logging is used to store progress information or
    time series data, as captured by the optimization
    :class:`~moptipy.api.process.Process` and activated by, e.g., the methods
    :meth:`~moptipy.api.execution.Execution.set_log_improvements` and
    :meth:`~moptipy.api.execution.Execution.set_log_all_fes`. It will look
    like this:

    >>> with InMemoryLogger() as logger:
    ...     with logger.csv("CSV", ["A", "B", "C"]) as csv:
    ...         csv.row((1, 2, 3))
    ...         csv.row([1.5, 2.0, 3.5])
    ...     print(logger.get_log())
    ['BEGIN_CSV', 'A;B;C', '1;2;3', '1.5;2;3.5', 'END_CSV']
    """

    def __init__(self, title: str, logger: Logger, header: list[str]) -> None:
        """
        Perform internal construction. Do not call directly.

        :param title: the title
        :param logger: the owning logger
        :param header: the header
        """
        super().__init__(title, logger)

        self.__header_len: Final[int] = len(header)
        if self.__header_len <= 0:
            # noinspection PyProtectedMember
            logger._error(f"Empty header {header} invalid for a CSV section")

        for c in header:
            if (not (isinstance(c, str))) or CSV_SEPARATOR in c:
                # noinspection PyProtectedMember
                logger._error(f"Invalid column {c}")

        # noinspection PyProtectedMember
        logger._write(CSV_SEPARATOR.join(
            [c.strip() for c in header]) + "\n")

    def row(self, row: tuple[int | float | bool, ...]
            | list[int | float | bool]) -> None:
        """
        Write a row of csv data.

        :param row: the row of data
        """
        if self.__header_len != len(row):
            # noinspection PyProtectedMember
            self._logger._error(
                f"Header of CSV section demands {self.__header_len} columns, "
                f"but row {row} has {len(row)}")

        # noinspection PyProtectedMember
        txt = [str(c) if isinstance(c, int)
               else bool_to_str(c) if isinstance(c, bool)  # type: ignore
               else (float_to_str(c) if isinstance(c, float) else
               cast(None, self._logger._error(
                   f"Invalid log value {c} in row {row}")))
               for c in row]

        # noinspection PyProtectedMember
        self._logger._write(f"{CSV_SEPARATOR.join(txt)}\n")


class KeyValueLogSection(LogSection):
    """
    A logger for key-value pairs.

    The key-values section `XXX` starts with the line `BEGIN_XXX` and ends
    with the line `END_XXX`. On every line in between, there is a key-value
    pair of the form `key: value`. Key-values sections support so-called
    scopes. Key-values pairs belong to a scope `Y` if the key starts with `Y.`
    followed by the actual key, e.g., `a.g: 5` denotes that the key `g` of
    scope `a` has value `5`. Such scopes can be arbitrarily nested: The
    key-value pair `x.y.z: 2` denotes a key `z` in the scope `y` nested within
    scope `x` having the value `2`. This system of nested scopes allows you
    to recursively invoke the method
    :meth:`~moptipy.api.component.Component.log_parameters_to` without
    worrying of key clashes. Just wrap the call to the `log_parameters_to`
    method of a sub-component into a unique scope. At the same time, this
    avoids the need of any more complex hierarchical data structures in our
    log files.

    >>> with InMemoryLogger() as logger:
    ...     with logger.key_values("A") as kv:
    ...         kv.key_value("x", 1)
    ...         with kv.scope("b") as sc1:
    ...             sc1.key_value("i", "j")
    ...             with sc1.scope("c") as sc2:
    ...                 sc2.key_value("l", 5)
    ...         kv.key_value("y", True)
    ...     print(logger.get_log())
    ['BEGIN_A', 'x: 1', 'b.i: j', 'b.c.l: 5', 'y: T', 'END_A']
    """

    def __init__(self, title: str | None,
                 logger: Logger, prefix: str, done) -> None:
        """
        Perform internal construction, do not call directly.

        :param title: the section title, or `None` for nested scopes.
        :param logger: the owning logger
        :param prefix: the prefix
        :param done: the set of already done keys and prefixes
        """
        if not isinstance(prefix, str):
            raise type_error(prefix, "prefix", str)
        super().__init__(title=title, logger=logger)
        self._prefix: Final[str] = prefix
        self.__done: Callable
        if done is None:
            self.__done = is_new()
            self.__done(prefix)
        else:
            self.__done = done
            if not done(prefix):
                # noinspection PyProtectedMember
                logger._error(f"Prefix {prefix!r} already done")

    def key_value(self, key: str, value,
                  also_hex: bool = False) -> None:
        """
        Write a key-value pair.

        Given key `A` and value `B`, the line `A: B` will be added to the log.
        If `value` (`B`) happens to be a floating point number, the value will
        also be stored in hexadecimal notation (:meth:`float.hex`).

        :param key: the key
        :param value: the value
        :param also_hex: also store the value as hexadecimal version
        """
        key = self._prefix + sanitize_name(key)
        if not self.__done(key):
            # noinspection PyProtectedMember
            self._logger._error(f"Key {key!r} already used")

        the_hex = None
        if isinstance(value, float):
            txt = float_to_str(value)
            if isfinite(value) and (also_hex
                                    or (("e" in txt) or ("." in txt))):
                the_hex = float.hex(value)
        elif isinstance(value, bool):
            txt = bool_to_str(value)
        else:
            txt = str(value)
            if also_hex and isinstance(value, int):
                the_hex = hex(value)

        txt = KEY_VALUE_SEPARATOR.join([key, txt])
        txt = f"{txt}\n"

        if the_hex:
            tmp = KEY_VALUE_SEPARATOR.join(
                [key + KEY_HEX_VALUE, the_hex])
            txt = f"{txt}{tmp}\n"

        # noinspection PyProtectedMember
        self._logger._write(txt)

    def scope(self, prefix: str) -> "KeyValueLogSection":
        """
        Create a new scope for key prefixes.

        :class:`KeyValueLogSection` only allows you to store flat key-value
        pairs where each key must be unique. However, what do we do if we
        have two components of an algorithm that have parameters with the
        same name (key)?
        We can hierarchically nest :class:`KeyValueLogSection` sections via
        prefix scopes. The idea is as follows: If one component has
        sub-components, instead of invoking their
        :meth:`~moptipy.api.component.Component.log_parameters_to` methods
        directly, which could lead to key-clashes, it will create a
        :meth:`scope` for each one and then pass these scopes to their
        :meth:`~moptipy.api.component.Component.log_parameters_to`.
        Each scope basically appends a prefix and a "." to the keys.
        If the prefixes are unique, this ensures that all prefix+"."+keys are
        unique, too.

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("A") as kv:
        ...         kv.key_value("x", "y")
        ...         with kv.scope("b") as sc1:
        ...             sc1.key_value("x", "y")
        ...             with sc1.scope("c") as sc2:
        ...                 sc2.key_value("x", "y")
        ...         with kv.scope("d") as sc3:
        ...             sc3.key_value("x", "y")
        ...             with sc3.scope("c") as sc4:
        ...                sc4.key_value("x", "y")
        ...     print(l.get_log())
        ['BEGIN_A', 'x: y', 'b.x: y', 'b.c.x: y', 'd.x: y', 'd.c.x: y', \
'END_A']

        :param prefix: the key prefix
        :return: the new logger
        """
        return KeyValueLogSection(
            title=None, logger=self._logger,
            prefix=(prefix if (self._prefix is None) else
                    f"{self._prefix}{sanitize_name(prefix)}."),
            done=self.__done)


class TextLogSection(LogSection):
    """
    A logger for raw, unprocessed text.

    Such a log section is used to capture the raw contents of the solutions
    discovered by the optimization :class:`~moptipy.api.process.Process`. For
    this purpose, our system will use the method
    :meth:`~moptipy.api.space.Space.to_str` of the search and/or solution
    :class:`~moptipy.api.space.Space`.

    >>> with InMemoryLogger() as logger:
    ...     with logger.text("T") as txt:
    ...         txt.write("Hello World!")
    ...     print(logger.get_log())
    ['BEGIN_T', 'Hello World!', 'END_T']
    """

    def __init__(self, title: str, logger: Logger) -> None:
        """
        Perform internal construction. Do not call it directly.

        :param title: the title
        :param logger: the logger
        """
        super().__init__(title, logger)
        # noinspection PyProtectedMember
        self.write = self._logger._write  # type: ignore


def parse_key_values(lines: Iterable[str]) -> dict[str, str]:
    """
    Parse a :meth:`~moptipy.utils.logger.Logger.key_values` section's text.

    :param lines: the lines with the key-values pairs
    :return: the dictionary with the

    >>> from moptipy.utils.logger import InMemoryLogger
    >>> with InMemoryLogger() as l:
    ...     with l.key_values("B") as kv:
    ...         kv.key_value("a", "b")
    ...         with kv.scope("c") as kvc:
    ...             kvc.key_value("d", 12)
    ...             kvc.key_value("e", True)
    ...         kv.key_value("f", 3)
    ...     txt = l.get_log()
    >>> print(txt)
    ['BEGIN_B', 'a: b', 'c.d: 12', 'c.e: T', 'f: 3', 'END_B']
    >>> dic = parse_key_values(txt[1:5])
    >>> keys = list(dic.keys())
    >>> keys.sort()
    >>> print(keys)
    ['a', 'c.d', 'c.e', 'f']
    """
    if not isinstance(lines, Iterable):
        raise type_error(lines, "lines", Iterable)
    dct = {}
    for line in lines:
        splt = line.split(KEY_VALUE_SEPARATOR)
        if len(splt) != 2:
            raise ValueError(
                f"Two strings separated by {KEY_VALUE_SEPARATOR!r} "
                f"expected, but encountered {len(splt)} in {line!r}.")
        key = splt[0].strip()
        if len(key) <= 0:
            raise ValueError(f"Empty key encountered in {line!r}.")
        value = splt[1].strip()
        if len(value) <= 0:
            raise ValueError(f"Empty value encountered in {line!r}.")
        dct[key] = value

    return dct



"""
The implementation of the hill climbing algorithm with restarts `hcr`.

This algorithm basically works like the normal hill climber `hc`
(:class:`~moptipy.algorithms.so.hill_climber.HillClimber`), but it will
restart automatically if no move was successful for
`max_moves_without_improvement` iterative steps. It therefore maintains an
internal counter `count` which is set to zero at the beginning of each restart
and which is also set to zero again any time a move successfully improved the
best-so-far solution of the current restart. If a search move, i.e., an
application of the unary operator, yielded a new solution which is not better
than the best-so-far solution of the current restart, `count` is incremented.
If `count >= max_moves_without_improvement`, the algorithm begins a new
restart with a new random solution.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import check_int_range


# start book
class HillClimberWithRestarts(Algorithm1):
    """
    The stochastic hill climbing algorithm only accepts improving moves.

    In each step, a hill climber creates a modified copy `new_x` of the
    current best solution `best_x`. If `new_x` is better than `best_x`,
    it becomes the new `best_x`. Otherwise, it is discarded. If no
    improvement is made for `max_moves_without_improvement` steps, the
    algorithm restarts.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the hill climber with restarts to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate
        limit: Final[int] = self.max_moves_without_improvement

        while not should_terminate():  # Until we need to quit....
            self.op0.op0(random, best_x)  # Create random solution and
            best_f: int | float = evaluate(best_x)
            count: int = 0  # The counter of unsuccessful moves = 0.

            while not should_terminate():  # Until we need to quit...
                op1(random, new_x, best_x)  # new_x=neighbor of best_x
                new_f: int | float = evaluate(new_x)
                if new_f < best_f:  # new_x is _better_ than best_x?
                    best_f = new_f  # Store its objective value.
                    best_x, new_x = new_x, best_x  # Swap best and new.
                    count = 0  # Reset unsuccessful move counter.
                else:  # The move did not lead to an improvement!?
                    count += 1  # Increment unsuccessful move counter.
                    if count >= limit:  # Too many unsuccessful moves?
                        break  # Break inner loop, start again randomly.
# end book

    def __init__(self, op0: Op0, op1: Op1,
                 max_moves_without_improvement: int) -> None:
        """
        Create the hill climber.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param max_moves_without_improvement: the maximum number of
            moves without improvement before a restart
        """
        super().__init__(
            f"hcr_{max_moves_without_improvement}", op0, op1)
        #: the maximum moves without improvement
        self.max_moves_without_improvement: Final[int] = \
            check_int_range(
                max_moves_without_improvement,
                "max_moves_without_improvement", 1, 1_000_000_000_000)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("maxMovesWithoutImprovement",
                         self.max_moves_without_improvement)



"""Some fixed demo codes for the JSSP examples."""
import sys
import urllib.request as rq
from typing import Callable, Final, Iterable

import numpy as np
from matplotlib.figure import Figure  # type: ignore

from moptipy.examples.jssp.experiment import INSTANCES
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.gantt_space import GanttSpace
from moptipy.examples.jssp.instance import (
    Instance,
    compute_makespan_lower_bound,
)
from moptipy.examples.jssp.ob_encoding import OperationBasedEncoding
from moptipy.examples.jssp.plot_gantt_chart import (
    marker_lb,
    marker_makespan,
    plot_gantt_chart,
)
from moptipy.spaces.permutations import Permutations
from moptipy.utils.console import logger
from moptipy.utils.lang import Lang
from moptipy.utils.path import UTF8, Path
from moptipy.utils.plot_utils import cm_to_inch, create_figure, save_figure
from moptipy.utils.types import type_error


def demo_instance() -> Instance:
    """
    Get the demo instance we use.

    :return: the demo instance
    """
    attr: Final[str] = "_res_"
    if not hasattr(demo_instance, attr):
        setattr(demo_instance, attr, Instance.from_resource("demo"))
    return getattr(demo_instance, attr)


def demo_search_space() -> Permutations:
    """
    Obtain an instance of the demo search space.

    :return: the demo search space
    """
    attr: Final[str] = "_res_"
    if not hasattr(demo_search_space, attr):
        instance: Final[Instance] = demo_instance()
        setattr(demo_search_space, attr,
                Permutations.with_repetitions(instance.jobs,
                                              instance.machines))
    return getattr(demo_search_space, attr)


def demo_point_in_search_space(optimum: bool = False) -> np.ndarray:
    """
    Create a demo point in the search space.

    :param optimum: should we return the optimal solution?
    :return: the point
    """
    space: Final[Permutations] = demo_search_space()
    res = space.create()
    if optimum:
        np.copyto(res, [0, 1, 2, 3, 1, 0, 1, 2, 0, 1,
                        3, 2, 2, 3, 1, 3, 0, 2, 3, 0])
    else:
        np.copyto(res, [0, 2, 3, 2, 2, 3, 1, 1, 0, 3,
                        1, 3, 2, 1, 3, 2, 0, 1, 0, 0])
    space.validate(res)
    return res


def demo_solution_space() -> GanttSpace:
    """
    Obtain an instance of the demo solution space.

    :return: the demo solution space
    """
    attr: Final[str] = "_res_"
    if not hasattr(demo_solution_space, attr):
        instance: Final[Instance] = demo_instance()
        setattr(demo_solution_space, attr, GanttSpace(instance))
    return getattr(demo_solution_space, attr)


def demo_encoding() -> OperationBasedEncoding:
    """
    Obtain an instance of the demo encoding.

    :return: the demo encoding
    """
    attr: Final[str] = "_res_"
    if not hasattr(demo_encoding, attr):
        instance: Final[Instance] = demo_instance()
        setattr(demo_encoding, attr, OperationBasedEncoding(instance))
    return getattr(demo_encoding, attr)


def demo_solution(optimum: bool = False) -> Gantt:
    """
    Create a demo solution.

    :param optimum: should we return the optimal solution?
    :return: the demo solution
    """
    space: Final[GanttSpace] = demo_solution_space()
    result: Final[Gantt] = space.create()
    demo_encoding().decode(demo_point_in_search_space(optimum=optimum),
                           result)
    space.validate(result)
    return result


def __make_gantt_demo_name(optimum: bool,
                           with_makespan: bool,
                           with_lower_bound: bool) -> str:
    """
    Construct the name for the demo gantt chart.

    :param optimum: should we return the optimal solution?
    :param with_makespan: should the makespan be included?
    :param with_lower_bound: should the lower bound be included?
    :return: the file name
    """
    prefix: str = "gantt_demo_opt_" if optimum else "gantt_demo_"
    if with_makespan:
        if with_lower_bound:
            return prefix + "with_makespan_and_lb"
        return prefix + "with_makespan"
    if with_lower_bound:
        return prefix + "with_lb"
    return prefix + "without_makespan"


def demo_gantt_chart(dirname: str,
                     optimum: bool = False,
                     with_makespan: bool = False,
                     with_lower_bound: bool = False,
                     width: float | int | None = cm_to_inch(10),
                     height: float | int | None = None,
                     filename: str | Callable =
                     __make_gantt_demo_name) -> list[Path]:
    """
    Plot the demo gantt chart.

    :param dirname: the directory
    :param optimum: should we return the optimal solution?
    :param with_makespan: should the makespan be included?
    :param with_lower_bound: should the lower bound be included?
    :param width: the optional width
    :param height: the optional height
    :param filename: the file name
    """
    the_dir: Final[Path] = Path.path(dirname)
    the_dir.ensure_dir_exists()

    if callable(filename):
        filename = filename(optimum,
                            with_makespan,
                            with_lower_bound)
    if not isinstance(filename, str):
        raise type_error(filename, "filename", str)

    result: Final[list[Path]] = []
    markers: list[Callable] = []

    if with_makespan:
        def info(g: Gantt) -> str:
            return Lang.current().format_str("gantt_info", gantt=g)
        if not optimum:
            markers.append(marker_makespan)
    else:
        def info(g: Gantt) -> str:
            return Lang.current().format_str("gantt_info_no_ms", gantt=g)

    if with_lower_bound:
        markers.append(marker_lb)

    for lang in Lang.all_langs():
        lang.set_current()
        figure: Figure = create_figure(width=width, height=height)

        gantt = demo_solution(optimum=optimum)
        plot_gantt_chart(gantt=gantt,
                         figure=figure,
                         x_label_inside=False,
                         y_label_inside=False,
                         markers=markers,
                         info=info)

        result.extend(save_figure(fig=figure,
                                  dir_name=the_dir,
                                  file_name=lang.filename(filename),
                                  formats="svg"))
    return result


def makespan_lower_bound_table(
        dirname: str,
        filename: str = "makespan_lower_bound",
        instances: Iterable[str] =
        tuple(["demo"] + list(INSTANCES))) -> Path:  # noqa
    """
    Make a table with the makespan lower bounds.

    Larger lower bounds are taken from the repository
    https://github.com/thomasWeise/jsspInstancesAndResults.

    :param dirname: the directory where to store the generated file.
    :param filename: the filename
    :param instances: the instances
    :returns: the generated file
    """
    insts = list(instances)
    insts.sort()
    i = insts.index("demo")
    if i >= 0:
        del insts[i]
        insts.insert(0, "demo")

    # get the data with the lower bounds information
    url: Final[str] = "https://github.com/thomasWeise/jsspInstancesAndResults"
    logger(f"now loading data from {url!r}.")
    with rq.urlopen(url) as f:  # nosec
        data: str = f.read().decode(UTF8).strip()
    if not data:
        raise ValueError(f"Could not load data form {url}.")
    logger(f"finished loading data from {url!r}.")
    start = data.find("<table>")
    if start <= 0:
        raise ValueError(f"Could not find <table> in {data}.")
    end = data.rfind("</dl>", start)
    if end <= start:
        raise ValueError(f"Could not find </dl> in {data}.")
    data = data[start:end].strip()

    lang: Final[Lang] = Lang.current()

    text: Final[list[str]] = [
        r"|name|$\jsspJobs$|$\jsspMachines$|$\lowerBound(\objf)$|"
        + r"$\lowerBound(\objf)^{\star}$|source for&nbsp;"
        + r"$\lowerBound(\objf)^{\star}$|",
        "|:--|--:|--:|--:|--:|:--|",
    ]
    bsrc: Final[str] = "[@eq:jsspLowerBound]"

    # compute the data
    for instn in insts:
        solved_to_optimality: bool = False
        inst = Instance.from_resource(instn)
        lb = compute_makespan_lower_bound(machines=inst.machines,
                                          jobs=inst.jobs,
                                          matrix=inst)
        src = bsrc
        if instn == "demo":
            lbx = lb
            solved_to_optimality = True
        else:
            start = data.find(
                f'<td align="right"><strong>{instn}</strong></td>')
            if start <= 0:
                start = data.find(f'<td align="right">{instn}</td>')
                if start <= 0:
                    raise ValueError(f"Could not find instance {instn!r}.")
                prefix = ""
                suffix = "</td>"
                solved_to_optimality = True
            else:
                prefix = "<strong>"
                suffix = "</strong></td>"
            end = data.find("</tr>", start)
            if end <= start:
                raise ValueError(
                    f"Could not find end of instance {instn!r}.")
            sel = data[start:end].strip()
            logger(f"located instance text {sel!r} for {instn!r}.")
            prefix = f'<td align="right">{inst.jobs}</td>\n<td align="' \
                     f'right">{inst.machines}</td>\n<td align="right">' \
                     f"{prefix}"
            pi = sel.find(prefix)
            if pi <= 0:
                raise ValueError(
                    f"Could not find prefix for instance {instn!r}.")
            sel = sel[(pi + len(prefix)):]
            si = sel.find(suffix)
            if si <= 0:
                raise ValueError(
                    f"Could not find suffix for instance {instn!r}.")
            lbx = int(sel[0:si])
            npt = '<td align="center"><a href="#'
            npi = sel.find(npt, si)
            if npi < si:
                raise ValueError(
                    f"Could not find source start for instance {instn!r}.")
            sel = sel[npi + len(npt):]
            end = sel.find('"')
            if end <= 0:
                raise ValueError(
                    f"Could not find source end for instance {instn!r}.")
            srcname = sel[:end].strip()
            logger(f"Source name is {srcname}.")

            fsrc = data.find(f'<dt id="user-content-{srcname.lower()}">'
                             f"{srcname}</dt><dd>")
            if fsrc <= 0:
                raise ValueError(
                    f"Could not find source mark for instance {instn!r}.")
            fsrce = data.find("</dd>", fsrc)
            if fsrce <= fsrc:
                raise ValueError(
                    f"Could not find end mark for instance {instn!r}.")
            sel = data[fsrc:fsrce].strip()
            nm = sel.rfind("</a>")
            if nm <= 0:
                raise ValueError(
                    f"Could not find link end for instance {instn!r}"
                    f" and source name {srcname}.")
            fm = sel.rfind(">", 0, nm)
            if fm >= nm:
                raise ValueError(
                    f"Could not find label mark for instance {instn!r}"
                    f" and source name {srcname}.")
            src = f"[@{sel[(fm + 1):nm].strip()}]"

        if lbx <= lb:
            if lbx < lb:
                raise ValueError(f"Lbx {lbx} cannot be larger than lb {lb}.")
            if not solved_to_optimality:
                src = "[@eq:jsspLowerBound]"
        instname = f"`{instn}`"
        if not solved_to_optimality:
            instname = f"**{instname}**"

        text.append(
            f"|{instname}|{inst.machines}|{inst.jobs}|{lb}|{lbx}|{src}|")

    # write the result
    out_dir = Path.path(dirname)
    out_dir.ensure_dir_exists()
    out_file = out_dir.resolve_inside(lang.filename(filename) + ".md")
    out_file.write_all(text)
    out_file.enforce_file()
    logger(f"finished writing output results to file {out_file!r}.")
    return out_file


# are we being executed?
if __name__ == "__main__":
    dest_dir: Final[Path] = Path.path(sys.argv[1])
    dest_dir.ensure_dir_exists()
    logger(f"We will print the JSSP examples into dir {dest_dir!r}.")
    makespan_lower_bound_table(dest_dir)
    demo_gantt_chart(dest_dir, True)
    demo_gantt_chart(dest_dir, False)
    logger(f"Finished printing the JSSP examples into dir {dest_dir!r}.")



"""
The API for implementing optimization algorithms and -problems.

This package provides two things:

* the basic, abstract API for implementing optimization algorithms and
  problems
* the abstraction and implementation of black-box processes in form of
  the :mod:`~moptipy.api.process` API and its implementations

The former helps us to implement and plug together different components of
optimization problems and optimization algorithms. The latter allows us to
apply algorithms to problems and to collect the results in a transparent way.
It also permits logging of the algorithm progress or even the change of
dynamic parameters.

The `moptipy` API has the following major elements for implementing
optimization algorithms and their components. You are encouraged to read more
about the logic behind this structure in our free online book at
https://thomasweise.github.io/oa.

- :mod:`~moptipy.api.component` provides the base class
  :class:`~moptipy.api.component.Component` from which all components of
  algorithm and optimization problems inherit. It offers the basic methods
  :meth:`~moptipy.api.component.Component.log_parameters_to` for writing all
  parameters and configuration elements of a component to a
  :class:`~moptipy.utils.logger.KeyValueLogSection` of a log file (embodied by
  an instance of :class:`~moptipy.utils.logger.Logger`) and the method
  :meth:`~moptipy.api.component.Component.initialize`, which must be
  overwritten to do any initialization necessary before a run of the
  optimization process.
- :mod:`~moptipy.api.algorithm` provides the base class
  :class:`~moptipy.api.algorithm.Algorithm` from which any optimization
  algorithm will inherit and which defines the optimization algorithm
  API, most importantly the method
  :meth:`~moptipy.api.algorithm.Algorithm.solve` which must be overwritten to
  implement the algorithm behavior. Here you can also find subclasses such as
  :class:`~moptipy.api.algorithm.Algorithm0`,
  :class:`~moptipy.api.algorithm.Algorithm1`, and
  :class:`~moptipy.api.algorithm.Algorithm2` for algorithms that use nullary,
  nullary and unary, as well as nully, unary, and binary search operators.
- :mod:`~moptipy.api.encoding` provides the base class
  :class:`~moptipy.api.encoding.Encoding` which can be used to implement the
  decoding procedure :meth:`~moptipy.api.encoding.Encoding.decode` which
  translates elements of a search space to elements of the solution space.
  Such an encoding may be needed if the two spaces differ, as they do, for
  instance, in our Job Shop Scheduling Problem example (see
  :mod:`~moptipy.examples.jssp.ob_encoding`).
- :mod:`~moptipy.api.mo_algorithm` provides the base class
  :class:`~moptipy.api.mo_algorithm.MOAlgorithm` for multi-objective
  optimization algorithms, which is a subclass of
  :class:`~moptipy.api.algorithm.Algorithm`.
- :mod:`~moptipy.api.mo_archive` provides the base class
  :class:`~moptipy.api.mo_archive.MOArchivePruner` for pruning archives of
  non-dominated solutions during a multi-objective optimization process. Since
  multi-objective optimization problems may have many possible solutions, but
  we can only return/retain a finite number of them, it may be necessary to
  decide which solution to keep and which to dispose of. This is the task
  of the archive pruner.
- :mod:`~moptipy.api.objective` provides the base class
  :class:`~moptipy.api.objective.Objective` for objective functions, i.e., for
  implementing the criteria rating how good a solution is. Objective functions
  are subject to minimization and provide a method
  :meth:`~moptipy.api.objective.Objective.evaluate` which computes an integer
  or floating point value rating one solution. They may additionally have
  a :meth:`~moptipy.api.objective.Objective.lower_bound` and/or an
  :meth:`~moptipy.api.objective.Objective.upper_bound`.
- :mod:`~moptipy.api.mo_problem` provides the base class
  :class:`~moptipy.api.mo_problem.MOProblem`, which represents a
  multi-objective optimization problem. This is a subclass of
  :class:`~moptipy.api.objective.Objective` *and* at the same time, a
  collection of multiple instances of
  :class:`~moptipy.api.objective.Objective`. Each multi-objective problem can
  compute a vector of objective values via
  :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`, but also implements
  the single-objective :meth:`~moptipy.api.mo_problem.MOProblem.evaluate`
  method returning a default scalarization of the objective vector. This makes
  multi-objective optimization compatible with single-objective optimization.
  While actual multi-objective algorithms can work in a truly multi-objective
  fashion, logging can be unified and even single-objective methods can be
  applied. To allow for the representation of preferences, the default
  Pareto domination relationship can be overwritten by providing a custom
  implementation of :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`.
- :mod:`~moptipy.api.operators` provides the base classes
  :class:`~moptipy.api.operators.Op0`, :class:`~moptipy.api.operators.Op1`, and
  :class:`~moptipy.api.operators.Op2` fur nullary, unary, and binary search
  operators, respectively. These can be used to implement the methods
  :meth:`~moptipy.api.operators.Op0.op0`,
  :meth:`~moptipy.api.operators.Op1.op1`, and
  :meth:`~moptipy.api.operators.Op2.op2` that are used by metaheuristic
  optimization algorithms to sample solutions and to derive solutions from
  existing ones.
- :mod:`~moptipy.api.space` provides the base class
  :class:`~moptipy.api.space.Space` for implementing the functionality of
  search and solution spaces. An instance of :class:`~moptipy.api.space.Space`
  offers methods such as :meth:`~moptipy.api.space.Space.create` for creating
  a data structure for holding point in the search space (with undefined
  contents), :meth:`~moptipy.api.space.Space.copy` for copying one data
  structure to another one, :meth:`~moptipy.api.space.Space.to_str` and
  :meth:`~moptipy.api.space.Space.from_str` to convert a data structure to
  and from a string, :meth:`~moptipy.api.space.Space.is_equal` to check
  whether one data structure equals another one, and
  :meth:`~moptipy.api.space.Space.validate` to verify whether the contents
  of a data structure are valid. With these methods, optimization algorithms
  can create and copy the data structure containers to hold solutions as they
  need without requiring any information about the actual contents and
  layout of these structures. The search operators (see above) then can handle
  the actual processing of the data structures. At the same time, the string
  conversion routines allow for storing the results of algorithms in log
  files.

The algorithm and experiment execution API has the following major elements:

- :mod:`~moptipy.api.execution` provides the builder class
  :class:`~moptipy.api.execution.Execution` that is used to construct one
  application of an optimization algorithm to an optimization problem. It
  configures the termination criteria, what information should be logged, and
  which algorithm to apply to which problem. Its method
  :meth:`~moptipy.api.execution.Execution.execute` returns an instance of
  :class:`~moptipy.api.process.Process` that contains the state of the
  optimization *after* it is completed, i.e., after the algorithm was
  executed. Log files follow the specification at
  https://thomasweise.github.io/moptipy/#log-file-sections.
- The module :mod:`~moptipy.api.logging` holds mainly string constants that
  identify sections and keys for the data to be written to log files.
- :mod:`~moptipy.api.mo_execution` provides the builder class
  :class:`~moptipy.api.mo_execution.MOExecution`, which is the multi-objective
  equivalent of :class:`~moptipy.api.execution.Execution`.
- :mod:`~moptipy.api.experiment` offers the function
  :func:`~moptipy.api.experiment.run_experiment` which allows you to execute
  a structured experiment applying a set of optimization algorithms to a set
  of optimization problems in a reproducible fashion, in parallel or in a
  distributed way. It will create a folder structure as prescribed in
  https://thomasweise.github.io/moptipy/#file-names-and-folder-structure.
  The data from such a folder structure can then be read in by the experiment
  evaluation tools in package :mod:`~moptipy.evaluation` and discussed in
  https://thomasweise.github.io/moptipy/#evaluating-experiments.
- :mod:`~moptipy.api.process` provides the class
  :class:`~moptipy.api.process.Process`, which is the basis for all
  optimization processes. It offers the functionality of an objective
  function, the search space, the termination criterion, and the random
  number generator to the optimization algorithm. At the same time, it
  collects the best-so-far solution and writes log files (if needed) for the
  user.
- :mod:`~moptipy.api.mo_process` provides the class
  :class:`~moptipy.api.mo_process.MOProcess`, which is the multi-objective
  equivalent of :class:`~moptipy.api.process.Process`. It also collects an
  archive of solutions.
- :mod:`~moptipy.api.subprocesses` offers several functions to slice off
  computational budget of a process to allow an algorithm to use a
  sub-algorithm, to wrap processes for running external algorithms that do
  not respect the `moptipy` API, or to run sub-algorithms starting with a
  specified initial solution.
"""



"""Statistic runs are time-depending statistics over several runs."""
from dataclasses import dataclass
from math import erf, sqrt
from typing import Any, Callable, Final, Iterable

import numba  # type: ignore
import numpy as np

import moptipy.evaluation.statistics as statn
from moptipy.evaluation.base import MultiRun2DData, MultiRunData, PerRunData
from moptipy.evaluation.progress import Progress
from moptipy.utils.nputils import DEFAULT_FLOAT, DEFAULT_INT, is_np_float
from moptipy.utils.types import type_error

#: The value of the CDF of the standard normal distribution CDF at -1,
#: which corresponds to "mean - 1 * sd".
__Q159: Final[float] = (1.0 + erf(-1.0 / sqrt(2.0))) / 2.0

#: The value of the CDF of the standard normal distribution CDF at +1,
#: which corresponds to "mean + 1 * sd".
__Q841: Final[float] = (1.0 + erf(1.0 / sqrt(2.0))) / 2.0


def _unique_floats_1d(data: list[np.ndarray]) -> np.ndarray:
    """
    Get all unique values that are >= than the minimum of all arrays.

    :param data: the data
    :return: the `ndarray` with the sorted, unique values
    """
    res: np.ndarray = np.unique(np.concatenate(data).astype(DEFAULT_FLOAT))
    mini = res[0]  # old version: int = -9223372036854775808
    for d in data:
        mini2 = d[0]
        if d[0] > mini:
            mini = mini2
    i: Final[int] = int(np.searchsorted(res, mini))
    if i > 0:
        return res[i:]
    return res


def __apply_fun(x_unique: np.ndarray,
                x_raw: list[np.ndarray],
                y_raw: list[np.ndarray],
                stat_func: Callable,
                out_len: int,
                dest_y: np.ndarray,
                stat_dim: int,
                values_buf: np.ndarray,
                pos_buf: np.ndarray) -> np.ndarray:
    """
    Perform the work of computing the time-depending statistic.

    The unique x-values `x_unique` have separately been computed with
    :func:`_unique_floats_1d` from `x_raw` so that they can be reused.
    `x_raw` and `y_raw` are lists with the raw time and objective data,
    respectively. `stat_fun` is the statistic function that will be applied to
    the step-wise generated data filled into `values_buf`. `pos_buf` will be
    used maintain the current indices into `x_raw` and `y_raw`. `dest_y` will
    be filled with the computed statistic for each element of `x_unique`.
    In a final step, we will remove all redundant elements of both arrays: If
    `x_unique` increases but `dest_y` remains the same, then the corresponding
    point is deleted if it is not the last point in the list. As a result,
    a two-dimensional time/value array is returned.

    :param x_unique: the unique time coordinates
    :param x_raw: a tuple of several x-data arrays
    :param y_raw: a tuple of several y-data arrays
    :param Callable stat_func: a statistic function which must have been
        jitted with numba
    :param out_len: the length of `dest_y` and `x_unique`
    :param dest_y: the destination array for the computed statistics
    :param stat_dim: the dimension of the tuples `x_raw` and `y_raw`
    :param values_buf: the buffer for the values to be passed to `stat_func`
    :param pos_buf: the position buffer
    :return: the two-dimensional `np.ndarray` where the first column is the
        time and the second column is the statistic value
    """
    for i in range(out_len - 1, -1, -1):  # reverse iteration
        x = x_unique[i]  # x_unique holds all unique x values
        for j in range(stat_dim):  # for all Progress datasets do
            idx = pos_buf[j]  # get the current position
            if x < x_raw[j][idx]:  # if x < then current time value
                idx = idx - 1  # step back by one
                pos_buf[j] = idx  # now x >= x_raw[j][idx]
            values_buf[j] = y_raw[j][idx]
        dest_y[i] = stat_func(values_buf)

    changes = 1 + np.flatnonzero(dest_y[1:] != dest_y[:-1])
    dest_len = len(dest_y) - 1
    changes_len = len(changes)
    if changes_len < 2:  # strange corner case: all values are the same
        # if there is only one value, use only that value
        # otherwise, use first and last value
        indexes = np.array([0]) if dest_len <= 1 else np.array([0, dest_len])
    elif changes[-1] != dest_len:  # always put last point
        indexes = np.concatenate((np.array([0]), changes,
                                  np.array([dest_len])))
    else:
        indexes = np.concatenate((np.array([0]), changes))
    return np.column_stack((x_unique[indexes], dest_y[indexes]))


def _apply_fun(x_unique: np.ndarray, x_raw: list[np.ndarray],
               y_raw: list[np.ndarray], stat_func: Callable) -> np.ndarray:
    """
    Compute a time-depending statistic.

    The unique x-values `x_unique` have separate been computed with
    `_unique_floats_1d` from `x_raw` so that they can be reused.
    `x_raw` and `y_raw` are tuples with the raw time and objective data,
    respectively. `stat_fun` is the statistic function that will be applied.
    In a final step, we will remove all redundant elements of both arrays: If
    `x_unique` increases but `dest_y` remains the same, then the corresponding
    point is deleted if it is not the last point in the list. As a result,
    a two-dimensional time/value array is returned. This function uses
    :meth:`__apply_fun` as internal work horse.

    :param x_unique: the unique time coordinates
    :param x_raw: a tuple of several x-data arrays
    :param y_raw: a tuple of several y-data arrays
    :param stat_func: a statistic function which must have been jitted with
        numba
    :return: the two-dimensional `numpy.ndarray` where the first column is the
        time and the second column is the statistic value
    """
    out_len: Final[int] = len(x_unique)
    dest_y: Final[np.ndarray] = np.zeros(out_len, DEFAULT_FLOAT)
    stat_dim: Final[int] = len(x_raw)
    values: Final[np.ndarray] = np.zeros(stat_dim, DEFAULT_FLOAT)
    pos: Final[np.ndarray] = np.array([len(x) - 1 for x in x_raw], DEFAULT_INT)

    return __apply_fun(x_unique, x_raw, y_raw, stat_func, out_len,
                       dest_y, stat_dim, values, pos)


@numba.njit(parallel=True)
def __stat_arith_mean(data: np.ndarray) -> np.number:
    """
    Compute the arithmetic mean.

    :param data: the data
    :return: the arithmetic mean
    """
    return data.mean()


@numba.njit(parallel=True)
def __stat_geo_mean(data: np.ndarray) -> np.number:
    """
    Compute the geometric mean.

    :param data: the data
    :return: the geometric mean
    """
    return np.exp(np.mean(np.log(data)))


@numba.njit(parallel=True)
def __stat_min(data: np.ndarray) -> np.number:
    """
    Compute the minimum.

    :param data: the data
    :return: the minimum
    """
    return data.min()


@numba.njit(parallel=True)
def __stat_max(data: np.ndarray) -> np.number:
    """
    Compute the maximum.

    :param data: the data
    :return: the maximum
    """
    return data.max()


@numba.njit
def __stat_median(data: np.ndarray) -> np.ndarray:
    """
    Compute the median.

    :param data: the data
    :return: the median
    """
    return np.median(data)


@numba.njit(parallel=True)
def __stat_sd(data: np.ndarray) -> np.number:
    """
    Compute the standard deviation.

    :param data: the data
    :return: the standard deviation
    """
    return data.std()


@numba.njit(parallel=True)
def __stat_mean_minus_sd(data: np.ndarray) -> np.number:
    """
    Compute the arithmetic mean minus the standard deviation.

    :param data: the data
    :return: the arithmetic mean minus the standard deviation
    """
    return data.mean() - data.std()


@numba.njit(parallel=True)
def __stat_mean_plus_sd(data: np.ndarray) -> np.number:
    """
    Compute the arithmetic mean plus the standard deviation.

    :param data: the data
    :return: the arithmetic mean plus the standard deviation
    """
    return data.mean() + data.std()


@numba.njit
def __stat_quantile_10(data: np.ndarray) -> np.ndarray:
    """
    Compute the 10% quantile.

    :param data: the data
    :return: the 10% quantile
    """
    length: Final[int] = len(data)
    if (length > 10) and ((length % 10) == 1):
        data.sort()
        return data[(length - 1) // 10]
    return np.quantile(data, 0.1)


@numba.njit
def __stat_quantile_90(data: np.ndarray) -> np.ndarray:
    """
    Compute the 90% quantile.

    :param data: the data
    :return: the 90% quantile
    """
    length: Final[int] = len(data)
    if (length > 10) and ((length % 10) == 1):
        data.sort()
        return data[(9 * (length - 1)) // 10]
    return np.quantile(data, 0.9)


@numba.njit
def __stat_quantile_159(data: np.ndarray) -> np.ndarray:
    """
    Compute the 15.9% quantile, which equals mean-sd in normal distributions.

    :param data: the data
    :return: the 15.9% quantile
    """
    return np.quantile(data, __Q159)


@numba.njit
def __stat_quantile_841(data: np.ndarray) -> np.ndarray:
    """
    Compute the 84.1% quantile, which equals mean+sd in normal distributions.

    :param data: the data
    :return: the 84.1% quantile
    """
    return np.quantile(data, __Q841)


#: The statistics key for the minimum
STAT_MINIMUM: Final[str] = statn.KEY_MINIMUM
#: The statistics key for the median.
STAT_MEDIAN: Final[str] = statn.KEY_MEDIAN
#: The statistics key for the arithmetic mean.
STAT_MEAN_ARITH: Final[str] = statn.KEY_MEAN_ARITH
#: The statistics key for the geometric mean.
STAT_MEAN_GEOM: Final[str] = statn.KEY_MEAN_GEOM
#: The statistics key for the maximum
STAT_MAXIMUM: Final[str] = statn.KEY_MAXIMUM
#: The statistics key for the standard deviation
STAT_STDDEV: Final[str] = statn.KEY_STDDEV
#: The key for the arithmetic mean minus the standard deviation.
STAT_MEAN_MINUS_STDDEV: Final[str] = f"{STAT_MEAN_ARITH}-{STAT_STDDEV}"
#: The key for the arithmetic mean plus the standard deviation.
STAT_MEAN_PLUS_STDDEV: Final[str] = f"{STAT_MEAN_ARITH}+{STAT_STDDEV}"
#: The key for the 10% quantile.
STAT_Q10: Final[str] = "q10"
#: The key for the 90% quantile.
STAT_Q90: Final[str] = "q90"
#: The key for the 15.9% quantile. In a normal distribution, this quantile
#: is where "mean - standard deviation" is located-
STAT_Q159: Final[str] = "q159"
#: The key for the 84.1% quantile. In a normal distribution, this quantile
#: is where "mean + standard deviation" is located-
STAT_Q841: Final[str] = "q841"

#: The internal function map.
_FUNC_MAP: Final[dict[str, Callable]] = {
    STAT_MINIMUM: __stat_min,
    STAT_MEDIAN: __stat_median,
    STAT_MEAN_ARITH: __stat_arith_mean,
    STAT_MEAN_GEOM: __stat_geo_mean,
    STAT_MAXIMUM: __stat_max,
    STAT_STDDEV: __stat_sd,
    STAT_MEAN_MINUS_STDDEV: __stat_mean_minus_sd,
    STAT_MEAN_PLUS_STDDEV: __stat_mean_plus_sd,
    STAT_Q10: __stat_quantile_10,
    STAT_Q90: __stat_quantile_90,
    STAT_Q159: __stat_quantile_159,
    STAT_Q841: __stat_quantile_841,
}


@dataclass(frozen=True, init=False, order=True)
class StatRun(MultiRun2DData):
    """A time-value statistic over a set of runs."""

    #: The name of this statistic.
    stat_name: str
    #: The time-dependent statistic.
    stat: np.ndarray

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 n: int,
                 time_unit: str,
                 f_name: str,
                 stat_name: str,
                 stat: np.ndarray):
        """
        Create the time-based statistics of an algorithm-setup combination.

        :param algorithm: the algorithm name, if all runs are
            with the same algorithm
        :param instance: the instance name, if all runs are
            on the same instance
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        :param stat_name: the name of the statistic
        :param stat: the statistic itself
        """
        super().__init__(algorithm, instance, n, time_unit, f_name)

        if not isinstance(stat_name, str):
            raise type_error(stat_name, "stat_name", str)
        object.__setattr__(self, "stat_name", stat_name)
        if not isinstance(stat, np.ndarray):
            raise type_error(stat, "statistic data", np.ndarray)
        stat.flags.writeable = False
        if (len(stat.shape) != 2) or (stat.shape[1] != 2) or \
                (stat.shape[0] <= 0):
            raise ValueError(
                "time array must be two-dimensional and have two columns and "
                f"at least one row, but has shape {stat.shape}.")
        if not is_np_float(stat.dtype):
            raise ValueError("statistics array must be float-typed, but has "
                             f"dtype {stat.dtype}.")
        object.__setattr__(self, "stat", stat)

    @staticmethod
    def create(source: Iterable[Progress],
               statistics: str | Iterable[str],
               consumer: Callable[["StatRun"], Any]) -> None:
        """
        Compute statistics from an iterable of `Progress` objects.

        :param source: the progress data
        :param statistics: the statistics to be computed
        :param consumer: the consumer for the statistics
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if isinstance(statistics, str):
            statistics = [statistics]
        if not isinstance(statistics, Iterable):
            raise type_error(statistics, "statistics", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)

        algorithm: str | None = None
        instance: str | None = None
        time_unit: str | None = None
        f_name: str | None = None
        time: list[np.ndarray] = []
        f: list[np.ndarray] = []
        n: int = 0

        for progress in source:
            if not isinstance(progress, Progress):
                raise type_error(progress, "stat run data source", Progress)
            if n <= 0:
                algorithm = progress.algorithm
                instance = progress.instance
                time_unit = progress.time_unit
                f_name = progress.f_name
            else:
                if algorithm != progress.algorithm:
                    algorithm = None
                if instance != progress.instance:
                    instance = None
                if time_unit != progress.time_unit:
                    raise ValueError(
                        f"Cannot mix time units {time_unit} "
                        f"and {progress.time_unit}.")
                if f_name != progress.f_name:
                    raise ValueError(f"Cannot mix f-names {f_name} "
                                     f"and {progress.f_name}.")
            n += 1
            time.append(progress.time)
            f.append(progress.f)

        if n <= 0:
            raise ValueError("Did not encounter any progress information.")

        x_unique = _unique_floats_1d(time)
        if not isinstance(x_unique, np.ndarray):
            raise type_error(x_unique, "x_unique", np.ndarray)
        if not is_np_float(x_unique.dtype):
            raise TypeError(
                f"x_unique must be floats, but is {x_unique.dtype}.")
        if (len(x_unique.shape) != 1) or (x_unique.shape[0] <= 0):
            raise ValueError(
                f"Invalid shape of unique values {x_unique.shape}.")

        count = 0
        for name in statistics:
            if not isinstance(name, str):
                raise type_error(name, "statistic name", str)
            if name not in _FUNC_MAP:
                raise ValueError(f"Unknown statistic name {name!r}.")
            consumer(StatRun(algorithm, instance, n, time_unit, f_name, name,
                             _apply_fun(x_unique, time, f, _FUNC_MAP[name])))
            count += 1

        if count <= 0:
            raise ValueError("No statistic names provided.")

    @staticmethod
    def from_progress(source: Iterable[Progress],
                      statistics: str | Iterable[str],
                      consumer: Callable[["StatRun"], Any],
                      join_all_algorithms: bool = False,
                      join_all_instances: bool = False) -> None:
        """
        Aggregate statist runs over a stream of progress data.

        :param source: the stream of progress data
        :param statistics: the statistics that should be computed per group
        :param consumer: the destination to which the new stat runs will be
            passed, can be the `append` method of a :class:`list`
        :param join_all_algorithms: should the statistics be aggregated
            over all algorithms
        :param join_all_instances: should the statistics be aggregated
            over all algorithms
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if isinstance(statistics, str):
            statistics = [statistics]
        if not isinstance(statistics, Iterable):
            raise type_error(statistics, "statistics", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(join_all_algorithms, bool):
            raise type_error(join_all_algorithms, "join_all_algorithms", bool)
        if not isinstance(join_all_instances, bool):
            raise type_error(join_all_instances, "join_all_instances", bool)

        sorter: dict[str, list[Progress]] = {}
        for prog in source:
            if not isinstance(prog, Progress):
                raise type_error(prog, "progress source", Progress)
            a: str = "" if join_all_algorithms else prog.algorithm
            i: str = "" if join_all_instances else prog.instance
            key: str = f"{a}/{i}/{prog.time_unit}/{prog.f_name}"
            if key in sorter:
                lst = sorter[key]
            else:
                lst = []
                sorter[key] = lst
            lst.append(prog)

        if len(sorter) <= 0:
            raise ValueError("source must not be empty")

        if len(sorter) > 1:
            keys = list(sorter.keys())
            keys.sort()
            for key in keys:
                StatRun.create(sorter[key], statistics, consumer)
        else:
            StatRun.create(next(iter(sorter.values())), statistics, consumer)


def get_statistic(obj: PerRunData | MultiRunData) -> str | None:
    """
    Get the statistic of a given object.

    :param obj: the object
    :return: the statistic string, or `None` if no statistic is specified
    """
    return obj.stat_name if isinstance(obj, StatRun) else None



"""Functions that can be used to test component implementations."""
from typing import Final

from moptipy.api import logging
from moptipy.api.component import Component
from moptipy.utils.logger import (
    KEY_VALUE_SEPARATOR,
    SECTION_END,
    SECTION_START,
    InMemoryLogger,
)
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error, type_name_of


def validate_component(component: Component) -> None:
    """
    Check whether an object is a valid moptipy component.

    This test checks the conversion to string and the logging of parameters.
    This method must be called before the other `validate_*` methods provided
    in this package.

    :param component: the component to test
    :raises ValueError: if `component` is not a valid
        :class:`~moptipy.api.component.Component` instance
    :raises TypeError: if a type is wrong or `component` is not even an
        instance of :class:`~moptipy.api.component.Component`
    """
    if not isinstance(component, Component):
        raise type_error(component, "component", Component)
    if component.__class__ == Component:
        raise ValueError(
            "component cannot be an instance of Component directly.")

    name = str(component)
    if not isinstance(name, str):
        raise type_error(name, "str(component)", str)
    if len(name) <= 0:
        raise ValueError("str(component) must return a non-empty string, "
                         f"but returns a {name!r}.")
    if name.strip() != name:
        raise ValueError("str(component) must return a string without "
                         "leading or trailing white space, "
                         f"but returns a {name!r}.")

    clean_name = sanitize_name(name)
    if clean_name != name:
        raise ValueError(
            "str(component) must return a string which does not "
            f"change when being sanitized, but returned {name!r},"
            f" which becomes {clean_name!r}.")
    name = str(component)
    if clean_name != name:
        raise ValueError("str(component) must always return the same value, "
                         f"but returns a {name!r} and {clean_name!r}.")

    name = repr(component)
    if name != clean_name:
        raise ValueError("repr(component) must equal str(component), but "
                         f"got {clean_name!r} vs. {name!r}.")

    if not (hasattr(component, "initialize")
            and callable(getattr(component, "initialize"))):
        raise ValueError("component must have method initialize.")
    component.initialize()

    if name != str(component):
        raise ValueError(f"name changed to {str(component)!r} "
                         f"from {name!r} after initialize!")

    # test the logging of parameter values
    if not (hasattr(component, "log_parameters_to")
            and callable(getattr(component, "log_parameters_to"))):
        raise ValueError("component must have method log_parameters_to.")

    secname: Final[str] = "KV"
    with InMemoryLogger() as log:
        with log.key_values(secname) as kv:
            component.log_parameters_to(kv)
        lines = log.get_log()

    ll = len(lines) - 1
    if (lines[0] != SECTION_START + secname) or \
            (lines[ll] != SECTION_END + secname):
        raise ValueError("Invalid log data produced '"
                         + "\n".join(lines) + "'.")

    kvs: Final[str] = KEY_VALUE_SEPARATOR
    lines = lines[1:ll]
    if len(lines) < 2:
        raise ValueError("A component must produce at least two lines of "
                         f"key-value data, but produced {len(lines)}.")

    done_keys: set[str] = set()
    idx: int = 0
    key: str = logging.KEY_NAME
    done_keys.add(key)
    keystr: str = f"{key}{kvs}"
    line: str = lines[idx]
    idx += 1
    if not line.startswith(keystr):
        raise ValueError(
            f"First log line must begin with {keystr!r}, but starts"
            f" with {line!r}.")
    rest = line[len(keystr):]
    if rest != name:
        raise ValueError(
            f"value of key {keystr!r} should equal "
            f"{name!r} but is {rest!r}.")

    key = logging.KEY_CLASS
    keystr = f"{key}{kvs}"
    done_keys.add(key)
    line = lines[idx]
    idx += 1
    if not line.startswith(keystr):
        raise ValueError(
            f"Second log line must begin with {keystr!r}, but "
            f"starts with {line!r}.")
    rest = line[len(keystr):]
    want = type_name_of(component)
    if rest != want:
        raise ValueError(
            f"value of key {keystr!r} should equal "
            f"{want!r} but is {rest!r}.")

    for line in lines[idx:]:
        i = line.index(kvs)
        key = line[0:i].strip()
        b = line[i + len(kvs)].strip()
        if (len(key) <= 0) or (len(b) <= 0):
            raise ValueError(
                f"Invalid key-value pair {line!r} - "
                f"splits to {(key + kvs + b)!r}!")
        if key in done_keys:
            raise ValueError(f"key {key!r} appears twice!")
        done_keys.add(key)



"""Functions that can be used to test algorithm implementations."""
from math import inf, isfinite
from typing import Any, Callable, Final

from moptipy.api.algorithm import (
    Algorithm,
    Algorithm0,
    Algorithm1,
    Algorithm2,
    check_algorithm,
)
from moptipy.api.encoding import Encoding
from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.api.operators import check_op0, check_op1, check_op2
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.objective import validate_objective
from moptipy.tests.space import validate_space
from moptipy.utils.nputils import rand_seed_generate
from moptipy.utils.types import check_int_range, type_error


def validate_algorithm(algorithm: Algorithm,
                       solution_space: Space,
                       objective: Objective,
                       search_space: Space | None = None,
                       encoding: Encoding | None = None,
                       max_fes: int = 100,
                       required_result: int | float | None = None,
                       uses_all_fes_if_goal_not_reached: bool = True,
                       is_encoding_deterministic: bool = True,
                       post: Callable[[Algorithm, int], Any] | None = None) \
        -> None:
    """
    Check whether an algorithm follows the moptipy API specification.

    :param algorithm: the algorithm to test
    :param solution_space: the solution space
    :param objective: the objective function
    :param search_space: the optional search space
    :param encoding: the optional encoding
    :param max_fes: the maximum number of FEs
    :param required_result: the optional required result quality
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    :param is_encoding_deterministic: is the encoding deterministic?
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    :raises TypeError: if `algorithm` is not a
        :class:`~moptipy.api.algorithm.Algorithm` instance
    :raises ValueError: if `algorithm` does not behave like it should
    """
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm)
    if (post is not None) and (not callable(post)):
        raise type_error(post, "post", None, call=True)

    check_algorithm(algorithm)
    if isinstance(algorithm, Algorithm0):
        check_op0(algorithm.op0)
        if isinstance(algorithm, Algorithm1):
            check_op1(algorithm.op1)
            if isinstance(algorithm, Algorithm2):
                check_op2(algorithm.op2)

    validate_component(algorithm)
    validate_space(solution_space, None)
    validate_objective(objective, None, None)
    if encoding is not None:
        validate_encoding(encoding, None, None, None,
                          is_encoding_deterministic)
        validate_space(search_space, None)

    check_int_range(max_fes, "max_fes", 1, 1_000_000_000)

    exp = Execution()
    exp.set_algorithm(algorithm)
    exp.set_max_fes(max_fes)
    exp.set_solution_space(solution_space)
    exp.set_objective(objective)
    seed: Final[int] = rand_seed_generate()
    if not isinstance(seed, int):
        raise type_error(seed, "seed", int)
    if not (0 <= seed <= 18446744073709551615):
        raise ValueError(f"invalid seed={seed}.")
    exp.set_rand_seed(seed)
    if search_space is not None:
        exp.set_search_space(search_space)
        exp.set_encoding(encoding)

    lb: Final[int | float] = objective.lower_bound()
    if (not isfinite(lb)) and (lb != -inf):
        raise ValueError(f"objective lower bound cannot be {lb}"
                         f" for {algorithm} on objective {objective}.")
    ub = objective.upper_bound()
    if (not isfinite(ub)) and (ub != inf):
        raise ValueError(f"objective upper bound cannot be {ub}"
                         f" for {algorithm} on objective {objective}.")

    if required_result is not None:
        if not (lb <= required_result <= ub):
            raise ValueError(f"required result must be in [{lb},{ub}], "
                             f"for {algorithm} on {objective} but "
                             f"is {required_result}")
        if (not isfinite(required_result)) and (required_result != -inf):
            raise ValueError(f"required_result must not be {required_result} "
                             f"for {algorithm} on {objective}.")

    progress: Final[tuple[list[int | float], list[int | float]]] = \
        [], []  # the progrss lists
    evaluate: Final[Callable[[Any], int | float]] = objective.evaluate

    for index in range(2 if is_encoding_deterministic else 1):

        if is_encoding_deterministic:

            def __k(xy, ii=index, ev=evaluate, pp=progress) -> int | float:
                rr = ev(xy)
                pp[ii].append(rr)
                return rr

            objective.evaluate = __k  # type: ignore

        with exp.execute() as process:
            # re-raise any exception that was caught
            if hasattr(process, "_caught"):
                error = getattr(process, "_caught")
                if error is not None:
                    raise error
            # no exception? ok, let's check the data
            if not process.has_best():
                raise ValueError(f"The algorithm {algorithm} did not produce "
                                 f"any solution on {objective} and "
                                 f"seed {seed}.")

            if (not process.should_terminate()) \
                    and uses_all_fes_if_goal_not_reached:
                raise ValueError(f"The algorithm {algorithm} stopped "
                                 f"before hitting the termination "
                                 f"criterion on {objective} and seed {seed}.")

            consumed_fes: int = check_int_range(
                process.get_consumed_fes(), "consumed_fes", 1, max_fes)
            last_imp_fe: int = check_int_range(
                process.get_last_improvement_fe(),
                "last_improvement_fe", 1, consumed_fes)
            consumed_time: int = check_int_range(
                process.get_consumed_time_millis(), "consumed_time",
                0, 100_0000_000)
            last_imp_time: int = check_int_range(
                process.get_last_improvement_time_millis(),
                "last_improvement_time", 0, consumed_time)
            if lb != process.lower_bound():
                raise ValueError(
                    "Inconsistent lower bounds between process "
                    f"({process.lower_bound()}) and objective ({lb})"
                    f" for {algorithm} on {objective} and seed {seed}.")
            if ub != process.upper_bound():
                raise ValueError(
                    "Inconsistent upper bounds between process "
                    f"({process.upper_bound()}) and objective ({ub}) "
                    f" for {algorithm} on {objective} and seed {seed}.")

            res_f: float | int = process.get_best_f()
            if not isfinite(res_f):
                raise ValueError(f"Infinite objective value of result "
                                 f"for {algorithm} on {objective}.")
            if (res_f < lb) or (res_f > ub):
                raise ValueError(f"Objective value {res_f} outside of bounds "
                                 f"[{lb},{ub}] for {algorithm} on "
                                 f"{objective} and seed {seed}.")

            if (required_result is not None) and (res_f > required_result):
                raise ValueError(
                    f"Algorithm {algorithm} should find solution of "
                    f"quality {required_result} on {objective}, but got "
                    f"one of {res_f} and seed {seed}.")

            if res_f <= lb:
                if last_imp_fe != consumed_fes:
                    raise ValueError(
                        f"if result={res_f} is as good as lb={lb}, then "
                        f"last_imp_fe={last_imp_fe} must equal"
                        f" consumed_fe={consumed_fes} for {algorithm} on "
                        f"{objective} and seed {seed}.")
                if (10_000 + (1.05 * last_imp_time)) < consumed_time:
                    raise ValueError(
                        f"if result={res_f} is as good as lb={lb}, then "
                        f"last_imp_time={last_imp_time} must not be much less"
                        f" than consumed_time={consumed_time} for "
                        f"{algorithm} on {objective} and seed {seed}.")
            elif uses_all_fes_if_goal_not_reached \
                    and (consumed_fes != max_fes):
                raise ValueError(
                    f"if result={res_f} is worse than lb={lb}, then "
                    f"consumed_fes={consumed_fes} must equal "
                    f"max_fes={max_fes} for {algorithm} on {objective}"
                    f" and seed {seed}.")

            y = solution_space.create()
            process.get_copy_of_best_y(y)
            solution_space.validate(y)
            check_f = objective.evaluate(y)
            if check_f != res_f:
                raise ValueError(
                    f"Inconsistent objective value {res_f} from process "
                    f"compared to {check_f} from objective function for "
                    f"{algorithm} on {objective} and seed {seed}.")

            x: Any | None = None
            if search_space is not None:
                x = search_space.create()
                process.get_copy_of_best_x(x)
                search_space.validate(x)

            if encoding is not None:
                y2 = solution_space.create()
                encoding.decode(x, y2)
                solution_space.validate(y2)
                if is_encoding_deterministic \
                        and not solution_space.is_equal(y, y2):
                    raise ValueError(
                        f"error when mapping point in search space {x} to "
                        f"solution {y2}, because it should be {y} for "
                        f"{algorithm} on {objective} under "
                        f"encoding {encoding} and seed {seed}")

            if post is not None:
                post(algorithm, consumed_fes)

    objective.evaluate = evaluate  # type: ignore

    if is_encoding_deterministic and (progress[0] != progress[1]):
        raise ValueError(f"when applying algorithm {algorithm} to "
                         f"{objective} under encoding {encoding} twice "
                         f"with the same seed {seed} did lead to different "
                         f"runs!")



"""
The base classes for implementing search operators.

Nullary search operators are used to sample the initial starting points of the
optimization processes. They inherit from class
:class:`~moptipy.api.operators.Op0`. The pre-defined unit test routine
:func:`~moptipy.tests.op0.validate_op0` can and should be used to test all the
nullary operators that are implemented.

Unary search operators accept one point in the search space as input and
generate a new, similar point as output. They inherit from class
:class:`~moptipy.api.operators.Op1`. The pre-defined unit test routine
:func:`~moptipy.tests.op1.validate_op1` can and should be used to test all the
unary operators that are implemented.

The basic unary operators :class:`~moptipy.api.operators.Op1` have no
parameter telling them how much of the input point to change. They may do a
hard-coded number of modifications (as, e.g.,
:class:`~moptipy.operators.permutations.op1_swap2.Op1Swap2` does) or may
apply a random number of modifications (like
:class:`~moptipy.operators.permutations.op1_swapn.Op1SwapN`). There is a
sub-class of unary operators named
:class:`~moptipy.api.operators.Op1WithStepSize` where a parameter `step_size`
with a value from the closed interval `[0.0, 1.0]` can be supplied. If
`step_size=0.0`, such an operator should perform the smallest possible
modification and for `step_size=1.0`, it should perform the largest possible
modification.

Binary search operators accept two points in the search space as input and
generate a new point that should be similar to both inputs as output. They
inherit from class :class:`~moptipy.api.operators.Op2`. The pre-defined unit
test routine :func:`~moptipy.tests.op2.validate_op2` can and should be used to
test all the binary operators that are implemented.
"""
from typing import Any

from numpy.random import Generator

from moptipy.api.component import Component
from moptipy.utils.types import type_error


# start op0
class Op0(Component):
    """A base class to implement a nullary search operator."""

    def op0(self, random: Generator, dest) -> None:
        """
        Apply the nullary search operator to fill object `dest`.

        Afterwards `dest` will hold a valid point in the search space.
        Often, this would be a point uniformly randomly sampled from the
        search space, but it could also be the result of a heuristic or
        even a specific solution.

        :param random: the random number generator
        :param dest: the destination data structure
        """
        raise ValueError("Method not implemented!")
# end op0


def check_op0(op0: Any) -> Op0:
    """
    Check whether an object is a valid instance of :class:`Op0`.

    :param op0: the (supposed) instance of :class:`Op0`
    :return: the object `op0`
    :raises TypeError: if `op0` is not an instance of :class:`Op0`

    >>> check_op0(Op0())
    Op0
    >>> try:
    ...     check_op0('A')
    ... except TypeError as te:
    ...     print(te)
    op0 should be an instance of moptipy.api.operators.Op0 but is \
str, namely 'A'.
    >>> try:
    ...     check_op0(None)
    ... except TypeError as te:
    ...     print(te)
    op0 should be an instance of moptipy.api.operators.Op0 but is None.
    """
    if isinstance(op0, Op0):
        return op0
    raise type_error(op0, "op0", Op0)


# start op1
class Op1(Component):
    """A base class to implement a unary search operator."""

    def op1(self, random: Generator, dest, x) -> None:
        """
        Fill `dest` with a modified copy of `x`.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x: the source point in the search space
        """
        raise ValueError("Method not implemented!")
# end op1


def check_op1(op1: Any) -> Op1:
    """
    Check whether an object is a valid instance of :class:`Op1`.

    :param op1: the (supposed) instance of :class:`Op1`
    :return: the object
    :raises TypeError: if `op1` is not an instance of :class:`Op1`

    >>> check_op1(Op1())
    Op1
    >>> try:
    ...     check_op1('A')
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1 but is str, \
namely 'A'.
    >>> try:
    ...     check_op1(None)
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1 but is None.
    """
    if isinstance(op1, Op1):
        return op1
    raise type_error(op1, "op1", Op1)


# start op2
class Op2(Component):
    """A base class to implement a binary search operator."""

    def op2(self, random: Generator, dest, x0, x1) -> None:
        """
        Fill `dest` with a combination of `x0` and `x1`.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x0: the first source point in the search space
        :param x1: the second source point in the search space
        """
        raise ValueError("Method not implemented!")
# end op2


def check_op2(op2: Any) -> Op2:
    """
    Check whether an object is a valid instance of :class:`Op2`.

    :param op2: the (supposed) instance of :class:`Op2`
    :return: the object `op2`
    :raises TypeError: if `op2` is not an instance of :class:`Op2`

    >>> check_op2(Op2())
    Op2
    >>> try:
    ...     check_op2('A')
    ... except TypeError as te:
    ...     print(te)
    op2 should be an instance of moptipy.api.operators.Op2 but is str, \
namely 'A'.
    >>> try:
    ...     check_op2(None)
    ... except TypeError as te:
    ...     print(te)
    op2 should be an instance of moptipy.api.operators.Op2 but is None.
    """
    if isinstance(op2, Op2):
        return op2
    raise type_error(op2, "op2", Op2)


# start op1WithStepSize
class Op1WithStepSize(Op1):
    """A unary search operator with a step size."""

    def op1(self, random: Generator, dest, x, step_size: float = 0.0) -> None:
        """
        Copy `x` to `dest` but apply a modification with a given `step_size`.

        This operator is similar to :meth:`Op1.op1` in that it stores a
        modified copy of `x` into `dest`. The difference is that you can also
        specify how much that copy should be different: The parameter
        `step_size` can take on any value in the interval `[0.0, 1.0]`,
        including the two boundary values. A `step_size` of `0.0` indicates
        the smallest possible move (for which `dest` will still be different
        from `x`) and `step_size=1.0` will lead to the largest possible move.

        The `step_size` may be interpreted differently by different operators:
        Some may interpret it as an exact requirement and enforce steps of the
        exact specified size, see, for example module
        :mod:`~moptipy.operators.bitstrings.op1_flip_m`. Others might
        interpret it stochastically as an expectation. Yet others may
        interpret it as a goal step width and try to realize it in a best
        effort kind of way, but may also do smaller or larger steps if the
        best effort fails, see for example module
        :mod:`~moptipy.operators.permutations.op1_swap_exactly_n`.
        What all operators should, however, have in common is that at
        `step_size=0.0`, they should try to perform a smallest possible change
        and at `step_size=1.0`, they should try to perform a largest possible
        change. For all values in between, step sizes should grow with rising
        `step_size`. This should allow algorithms that know nothing about the
        nature of the search space or the operator's moves to still tune
        between small and large moves based on a policy which makes sense in a
        black-box setting.

        Every implementation of :class:`Op1WithStepSize` must specify a
        reasonable default value for this parameter ensure compatibility with
        :meth:`Op1.op1`. In this base class, we set the default to `0.0`.

        Finally, if a `step_size` value is passed in which is outside the
        interval `[0, 1]`, the behavior of this method is undefined. It may
        throw an exception or not. It may also enter an infinite loop.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x: the source point in the search space
        :param step_size: the step size parameter for the unary operator
        """
        raise ValueError("Method not implemented!")
# end op1WithStepSize


def check_op1_with_step_size(op1: Any) -> Op1WithStepSize:
    """
    Check whether an object is a valid instance of :class:`Op1WithStepSize`.

    :param op1: the (supposed) instance of :class:`Op1WithStepSize`
    :return: the object `op1`
    :raises TypeError: if `op1` is not an instance of :class:`Op1WithStepSize`

    >>> check_op1_with_step_size(Op1WithStepSize())
    Op1WithStepSize
    >>> try:
    ...     check_op1_with_step_size('A')
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is str, namely 'A'.
    >>> try:
    ...     check_op1_with_step_size(Op1())
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is moptipy.api.operators.Op1, namely 'Op1'.
    >>> try:
    ...     check_op1_with_step_size(None)
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is None.
    """
    if isinstance(op1, Op1WithStepSize):
        return op1
    raise type_error(op1, "op1", Op1WithStepSize)



"""The HTML text format driver."""

from io import TextIOBase
from typing import Final

from moptipy.utils.formatted_string import (
    NAN,
    NEGATIVE_INFINITY,
    NUMBER,
    POSITIVE_INFINITY,
    SPECIAL,
    TEXT,
)
from moptipy.utils.text_format import MODE_TABLE_HEADER, TextFormatDriver

#: the default border style
_BORDER: Final[str] = "1pt solid black"

#: special characters in HTML
SPECIAL_CHARS: Final[dict[str, str]] = {
    "\u2205": "&#x2205;",
    "\u221E": "&#x221e;",
    "-\u221E": "-&#x221e;",
    "inf": "&#x221E;",
    "-inf": "-&#x221E;",
    "nan": "&#x2205;",
    "\u03b1": "&#x3b1;",
    "\u2014": "&mdash;",
}


class HTML(TextFormatDriver):
    """The HTML text driver."""

    def begin_table_body(self, stream: TextIOBase, cols: str) -> None:
        """Write the beginning of the table body."""
        stream.write(
            f'<table style="border:{_BORDER};border-collapse:collapse">')

    def end_table_body(self, stream: TextIOBase, cols: str) -> None:
        """Write the ending of the table body."""
        stream.write("</tbody></table>")

    def begin_table_header(self, stream: TextIOBase, cols: str) -> None:
        """Begin the header of an HTML table."""
        stream.write("<thead>")

    def end_table_header(self, stream: TextIOBase, cols: str) -> None:
        """End the header of an HTML table."""
        stream.write("</thead><tbody>")

    def begin_table_row(self, stream: TextIOBase, cols: str,
                        section_index: int, row_index: int,
                        row_mode: int) -> None:
        """Begin a row in an HTML table."""
        stream.write("<tr>")

    def end_table_row(self, stream: TextIOBase, cols: str,
                      section_index: int, row_index: int) -> None:
        """End a row in a HTML table."""
        stream.write("</tr>")

    def begin_table_cell(self, stream: TextIOBase, cols: str,
                         section_index: int, row_index: int,
                         col_index: int, cell_mode: int) -> None:
        """Begin an HTML table cell."""
        style: str = cols[col_index]
        style = "center" if style == "c" \
            else "left" if style == "l" else "right"
        style = f"padding:3pt;text-align:{style}"
        if col_index < (len(cols) - 1):
            style = f"{style};border-right:{_BORDER}"
        if row_index <= 0 <= section_index:
            style = f"{style};border-top:{_BORDER}"
        cell: str = "th" if (cell_mode == MODE_TABLE_HEADER) else "td"
        stream.write(f'<{cell} style="{style}">')  # noqa: B028

    def end_table_cell(self, stream: TextIOBase, cols: str,
                       section_index: int, row_index: int,
                       col_index: int, cell_mode: int) -> None:
        """End an HTML table cell."""
        stream.write("</th>" if (cell_mode == MODE_TABLE_HEADER) else "</td>")

    def text(self, stream: TextIOBase, text: str, bold: bool, italic: bool,
             code: bool, mode: int) -> None:
        """Print a text string."""
        if len(text) <= 0:
            return
        styles: str = ""
        if bold:
            styles = "font-weight:bold;"
        if italic:
            styles += "font-style:italic;"
        if code:
            styles += "font-family:monospace;"
        if len(styles) > 0:
            stream.write(f'<span style="{styles[0:-1]}">')  # noqa: B028

        if mode == TEXT:
            stream.write(text)
        elif mode == NUMBER:
            i: int = text.find("e")
            if i < 0:
                i = text.find("E")
            if i > 0:
                stream.write(f"{text[:i]}&#xD7;10<sup>{text[i + 1:]}</sup>")
            else:
                stream.write(text)
        elif mode == NAN:
            stream.write("&#x2205;")
        elif mode == POSITIVE_INFINITY:
            stream.write("&#x221E;")
        elif mode == NEGATIVE_INFINITY:
            stream.write("-&#x221E;")
        elif mode == SPECIAL:
            s: Final[str] = str(text)
            if s not in SPECIAL_CHARS:
                raise ValueError(f"invalid special character: {s!r}")
            stream.write(SPECIAL_CHARS[s])
        else:
            raise ValueError(f"invalid mode {mode} for text {text!r}.")

        if len(styles) > 0:
            stream.write("</span>")

    def __str__(self):
        """
        Get the appropriate file suffix.

        :returns: the file suffix
        :retval 'html': always
        """
        return "html"

    @staticmethod
    def instance() -> "HTML":
        """
        Get the HTML format singleton instance.

        :returns: the singleton instance of the HTML format
        """
        attr: Final[str] = "_instance"
        func: Final = HTML.instance
        if not hasattr(func, attr):
            setattr(func, attr, HTML())
        return getattr(func, attr)



"""Some tools for optimization."""

from math import exp, log

import numba  # type: ignore


@numba.njit(cache=True)
def exponential_step_size(step_size: float,
                          min_steps: int, max_steps: int) -> int:
    """
    Translate a step size in `[0,1]` to an integer in `min_steps..max_steps`.

    The idea is that we would like to have step-size dependent operators of
    type :class:`~moptipy.api.operators.Op1WithStepSize`. These operators
    allow algorithms to tune the amount of change to be applied to a solution
    between `0` and `1` (inclusively). As described in the documentation of
    :meth:`~moptipy.api.operators.Op1WithStepSize.op1`, `0` means that the
    smallest possible change should be applied and `1` means that the largest
    possible change should be applied.

    Now for many search spaces, we need to translate this step size from
    `[0,1]` to an integer. For instance, if we have `n`-dimensional
    :mod:`~moptipy.spaces.bitstrings` as search space, then we can flip
    anything between `1` and `n` bits. Straightforwardly, one would linearly
    scale the step size from `[0,1]` to `1..n`. Unfortunately, if we do that,
    then different values of `step_size` will have very different meaning
    depending on `n`. For example, `step_size=0.05` would translate to
    `round(1 + step_size * (n-1)) = 1` bits to be flipped if `n=10`, to
    `6` bit flips for `n=100`, and to `501` bit flips for `n=10_000`. While
    flipping one bit is a very small move and flipping six bits may be a
    medium-size move in discrete optimization, flipping over 500 bits is
    actually always quite a lot.

    What we would like is to have scale-independent small moves but still be
    able to make large moves. We can get this by exponentially transforming
    the `step_width`. Most `step_size` values will result in small integer
    steps and only `step_size` values close to `1` will yield really big
    results.

    And for a `step_size=0.05`, we get for different `n`:

    >>> exponential_step_size(0.05, 1, 10)
    1
    >>> exponential_step_size(0.05, 1, 100)
    1
    >>> exponential_step_size(0.05, 1, 10_000)
    2
    >>> exponential_step_size(0.05, 1, 1_000_000)
    2
    >>> exponential_step_size(0.05, 1, 1_000_000_000)
    3
    >>> exponential_step_size(0.05, 1, 1_000_000_000_000)
    4

    For different values of `step_size` and a fixed `n`, we can still obtain
    the whole spectrum of possible changes. For `n=10`, for example, we get:

    >>> exponential_step_size(0.0, 1, 10)
    1
    >>> exponential_step_size(0.1, 1, 10)
    1
    >>> exponential_step_size(0.2, 1, 10)
    2
    >>> exponential_step_size(0.3, 1, 10)
    2
    >>> exponential_step_size(0.4, 1, 10)
    3
    >>> exponential_step_size(0.5, 1, 10)
    3
    >>> exponential_step_size(0.6, 1, 10)
    4
    >>> exponential_step_size(0.7, 1, 10)
    5
    >>> exponential_step_size(0.8, 1, 10)
    6
    >>> exponential_step_size(0.85, 1, 10)
    7
    >>> exponential_step_size(0.9, 1, 10)
    8
    >>> exponential_step_size(0.95, 1, 10)
    9
    >>> exponential_step_size(1.0, 1, 10)
    10

    So we can still reach the whole range of possible steps from `1` to `n`.

    >>> isinstance(exponential_step_size(0.5, 1, 9), int)
    True
    >>> exponential_step_size(0.0, 1, 100)
    1
    >>> exponential_step_size(1.0, 1, 100)
    100
    >>> exponential_step_size(1.0 / 3.0, 1, 10)
    2
    >>> exponential_step_size(1.0 / 3.0, 1, 100)
    5
    >>> exponential_step_size(1.0 / 3.0, 1, 10_000)
    22
    >>> exponential_step_size(0.0, 2, 10)
    2
    >>> exponential_step_size(0.0, 9, 10)
    9
    >>> exponential_step_size(0.0, 10, 10)
    10
    >>> exponential_step_size(1.0, 10, 10)
    10

    :param step_size: the step size from `[0,1]` to be transformed to an
        integer
    :param min_steps: the minimum (inclusive) value for the returned integer
    :param max_steps: the maximum (inclusive) value for the returned integer
    """
    return round(min_steps + exp(step_size * log(
        max_steps - min_steps + 1))) - 1


@numba.njit(cache=True)
def inv_exponential_step_size(int_val: int,
                              min_steps: int, max_steps: int) -> float:
    """
    Compute the inverse of :func:`exponential_step_size`.

    This routine exists mainly to make testing easier.

    :param int_val: the return value of :func:`exponential_step_size`.
    :param min_steps: the minimum (inclusive) value any `int_val`
    :param max_steps: the maximum (inclusive) value any `int_val`

    >>> exponential_step_size(0.47712125471966244, 1, 10)
    3
    >>> inv_exponential_step_size(3, 1, 10)
    0.47712125471966244
    >>> inv_exponential_step_size(1, 1, 10)
    0.0
    >>> inv_exponential_step_size(10, 1, 10)
    1.0
    >>> inv_exponential_step_size(33, 6, 673)
    0.5123088678224029
    >>> exponential_step_size(0.5123088678224029, 6, 673)
    33
    >>> inv_exponential_step_size(3, 3, 3)
    1.0
    >>> inv_exponential_step_size(3, 3, 10)
    0.0
    >>> inv_exponential_step_size(10, 3, 10)
    1.0
    """
    if int_val >= max_steps:
        return 1.0
    if int_val <= min_steps:
        return 0.0
    return log(int_val - min_steps + 1) / log(max_steps - min_steps + 1)



"""Utilities for multi-objective optimization."""

import numba  # type: ignore
import numpy as np


@numba.njit(cache=True)
def dominates(a: np.ndarray, b: np.ndarray) -> int:
    """
    Check if one objective vector dominates or is dominated by another one.

    :param a: the first objective vector
    :param b: the second objective value
    :returns: an integer value indicating the domination relationship
    :retval -1: if `a` dominates `b`
    :retval 1: if `b` dominates `a`
    :retval 2: if `b` equals `a`
    :retval 0: if `a` and `b` are mutually non-dominated, i.e., if neither `a`
        dominates `b` not `b` dominates `a` and `b` is also different from `a`

    >>> from numpy import array
    >>> dominates(array([1, 1, 1]), array([2, 2, 2]))
    -1
    >>> dominates(array([1.0, 1.0, 2.0]), array([2.0, 2.0, 1.0]))
    0
    >>> dominates(array([2, 2, 2]), array([1, 1, 1]))
    1
    >>> dominates(array([2, 2, 2]), array([2, 2, 2]))
    2
    """
    res: int = 0
    for i, av in enumerate(a):
        bv = b[i]
        if av < bv:
            if res == 0:
                res = -1
            elif res == 1:
                return 0
        elif bv < av:
            if res == 0:
                res = 1
            elif res == -1:
                return 0
    return 2 if res == 0 else res


@numba.njit(cache=True)
def lexicographic(a: np.ndarray, b: np.ndarray) -> int:
    """
    Compare two arrays lexicographically.

    :param a: the first objective vector
    :param b: the second objective value
    :returns: `-1` if `a` is lexicographically less than `b`, `1` if `b` is
        less than `a`, `0` otherwise
    :retval -1: if `a` is lexicographically less than `b`
    :retval 1: if `b` is lexicographically less than `a`
    :retval 2: if `b` equals `a`

    >>> from numpy import array
    >>> lexicographic(array([1, 1, 1]), array([2, 2, 2]))
    -1
    >>> lexicographic(array([1, 1, 1]), array([1, 1, 2]))
    -1
    >>> lexicographic(array([2, 2, 2]), array([1, 1, 1]))
    1
    >>> lexicographic(array([2, 2, 2]), array([2, 2, 1]))
    1
    >>> lexicographic(array([2, 2, 2]), array([2, 2, 2]))
    0
    """
    for i, f in enumerate(a):
        k = b[i]
        if f < k:
            return -1
        if f > k:
            return 1
    return 0



"""The base class for implementing multi-objective problems."""
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable

import numpy as np
from numpy import empty

from moptipy.api.logging import KEY_SPACE_NUM_VARS, SCOPE_OBJECTIVE_FUNCTION
from moptipy.api.mo_problem import MOProblem
from moptipy.api.mo_utils import dominates
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.math import try_int
from moptipy.utils.nputils import (
    KEY_NUMPY_TYPE,
    dtype_for_data,
    numpy_type_to_str,
)
from moptipy.utils.types import type_error


class BasicMOProblem(MOProblem):
    """
    The base class for implementing multi-objective optimization problems.

    This class allows to construct a simple python function for scalarizing
    a vector of objective values in its constructor and also determines the
    right datatype for the objective vectors.

    It therefore first obtains the type (integers or floats?) of the objective
    values as well as the bounds of the objective functions. This is used to
    determine the right numpy `dtype` for the objective vectors. We want to
    represent objective vectors as compact as possible and use an integer
    vector if possible.

    Once this information is obtained, we invoke a call-back function
    `get_scalarizer` which should return a python function that computes the
    scalarization result, i.e., the single scalar value representing the
    vector of objective values in single-objective optimization. This function
    must be monotonous. If the bounds are finite, it is applied to the vector
    of lower and upper bounds to get the lower and upper bounds of the
    scalarization result.

    Examples for implementing this class are
    class:`~moptipy.mo.problem.weighted_sum.WeightedSum` and
    :class:`~moptipy.mo.problem.weighted_sum.Prioritize`, which represent a
    multi-objective optimization problem either as weighted sum or by
    priorizing the objective value (via an internal weighted sum).
    """

    def __init__(self, objectives: Iterable[Objective],
                 get_scalarizer: Callable[
                     [bool, int, list[int | float],
                      list[int | float]],
                     Callable[[np.ndarray], int | float]] = None,
                 domination: Callable[[np.ndarray, np.ndarray], int] | None
                 = dominates) -> None:
        """
        Create the basic multi-objective optimization problem.

        :param objectives: the objective functions
        :param get_scalarizer: Create the function for scalarizing the
            objective values. This constructor receives as parameters a `bool`
            which is `True` if and only if all objective functions always
            return integers and `False` otherwise, i.e., if at least one of
            them may return a `float`, the length of the f-vectors, and lists
            with the lower and upper bounds of the objective functions. It can
            use this information to dynamically create and return the most
            efficient scalarization function.
        :param domination: a function reflecting the domination relationship
            between two vectors of objective values. It must obey the contract
            of :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`, which is
            the same as :func:`moptipy.api.mo_utils.dominates`, to which it
            defaults. `None` overrides nothing.
        """
        if not isinstance(objectives, Iterable):
            raise type_error(objectives, "objectives", Iterable)
        if not callable(get_scalarizer):
            raise type_error(get_scalarizer, "get_scalarizer", call=True)

        lower_bounds: Final[list[int | float]] = []
        upper_bounds: Final[list[int | float]] = []
        calls: Final[list[Callable[[Any], int | float]]] = []
        min_lower_bound: int | float = inf
        max_upper_bound: int | float = -inf

        # Iterate over all objective functions and see whether they are
        # integer-valued and have finite bounds and to collect the bounds.
        always_int: bool = True
        is_int: bool
        lb: int | float
        ub: int | float
        for objective in objectives:
            if not isinstance(objective, Objective):
                raise type_error(objective, "objective[i]", Objective)
            is_int = objective.is_always_integer()
            always_int = always_int and is_int
            calls.append(objective.evaluate)
            lb = objective.lower_bound()
            if isfinite(lb):
                if is_int:
                    if not isinstance(lb, int):
                        raise ValueError(
                            f"if is_always_integer() of objective {objective}"
                            " is True, then lower_bound() must be infinite or"
                            f" int, but is {lb}.")
                else:
                    lb = try_int(lb)
                if lb < min_lower_bound:
                    min_lower_bound = lb
            else:
                min_lower_bound = -inf
            lower_bounds.append(lb)
            ub = objective.upper_bound()
            if isfinite(ub):
                if is_int:
                    if not isinstance(ub, int):
                        raise ValueError(
                            f"if is_always_integer() of objective {objective}"
                            " is True, then upper_bound() must be infinite "
                            f"or int, but is {ub}.")
                else:
                    ub = try_int(ub)
                if ub > max_upper_bound:
                    max_upper_bound = ub
            else:
                max_upper_bound = inf
            if lb >= ub:
                raise ValueError(
                    f"lower_bound()={lb} of objective {objective} must "
                    f"be < than upper_bound()={ub}")
            upper_bounds.append(ub)

        n: Final[int] = len(calls)
        if n <= 0:
            raise ValueError("No objective function found!")

        use_lb: int | float = min_lower_bound
        use_ub: int | float = max_upper_bound
        if always_int:
            if isfinite(min_lower_bound) and isfinite(max_upper_bound):
                use_lb = min(min_lower_bound,
                             min_lower_bound - max_upper_bound)
                use_ub = max(max_upper_bound,
                             max_upper_bound - min_lower_bound)
            else:
                use_lb = -inf
                use_ub = inf

        # Based on the above findings, determine the data type:
        #: The data type of the objective vectors.
        #: If the objectives all always are integers and have known and finite
        #: bounds, then we can use the smallest possible integer type.
        #: This type will be large enough to allow computing "a - b" of any two
        #: objective values "a" and "b" without overflow.
        #: If they are at least integer-valued, we can use the largest integer
        #: type.
        #: If also this is not True, then we just use floating points.
        self.__dtype: Final[np.dtype] = dtype_for_data(
            always_int, use_lb, use_ub)
        #: The dimension of the objective space.
        self.__dimension: Final[int] = n

        #: the creator function for objective vectors
        self.f_create = lambda nn=n, dt=self.__dtype: empty(  # type: ignore
            nn, dt)  # type: ignore

        #: the holder for lower bounds
        self.__lower_bounds: Final[tuple[int | float, ...]] = \
            tuple(lower_bounds)
        #: the holder for upper bounds
        self.__upper_bounds: Final[tuple[int | float, ...]] = \
            tuple(upper_bounds)

        # set up the scalarizer
        self._scalarize: Final[Callable[[np.ndarray], int | float]] \
            = get_scalarizer(always_int, n, lower_bounds, upper_bounds)
        if not callable(self._scalarize):
            raise type_error(self._scalarize, "result of get_scalarizer",
                             call=True)

        # compute the scalarized bounds
        temp: np.ndarray | None = None
        lb = -inf
        if isfinite(min_lower_bound):
            temp = np.array(lower_bounds, dtype=self.__dtype)
            lb = self._scalarize(temp)
            if not isinstance(lb, int | float):
                raise type_error(lb, "computed lower bound", (int, float))
            if (not isfinite(lb)) and (lb > -inf):
                raise ValueError("non-finite computed lower bound "
                                 f"can only be -inf, but is {lb}.")
            lb = try_int(lb)
        #: the lower bound of this scalarization
        self.__lower_bound: Final[int | float] = lb

        ub = inf
        if isfinite(max_upper_bound):
            temp = np.array(upper_bounds, dtype=self.__dtype)
            ub = self._scalarize(temp)
            if not isinstance(ub, int | float):
                raise type_error(ub, "computed upper bound", (int, float))
            if (not isfinite(ub)) and (ub < inf):
                raise ValueError("non-finite computed upper bound "
                                 f"can only be inf, but is {ub}.")
            ub = try_int(ub)
        #: the upper bound of this scalarization
        self.__upper_bound: Final[int | float] = ub

        #: the internal objectives
        self.__calls: Final[tuple[
            Callable[[Any], int | float], ...]] = tuple(calls)
        #: the objective functions
        self._objectives = tuple(objectives)

        #: the internal temporary array
        self._temp: Final[np.ndarray] = self.f_create() \
            if temp is None else temp

        if domination is not None:
            if not callable(domination):
                raise type_error(domination, "domination", call=True)
            self.f_dominates = domination  # type: ignore

    def initialize(self) -> None:
        """Initialize the multi-objective problem."""
        super().initialize()
        for ff in self._objectives:
            ff.initialize()

    def f_dimension(self) -> int:
        """
        Obtain the number of objective functions.

        :returns: the number of objective functions
        """
        return self.__dimension

    def f_dtype(self) -> np.dtype:
        """
        Get the data type used in `f_create`.

        :returns: the data type used by
            :meth:`moptipy.api.mo_problem.MOProblem.f_create`.
        """
        return self.__dtype

    def f_evaluate(self, x, fs: np.ndarray) -> int | float:
        """
        Perform the multi-objective evaluation of a solution.

        :param x: the solution to be evaluated
        :param fs: the array to receive the objective values
        :returns: the scalarized objective values
        """
        for i, o in enumerate(self.__calls):
            fs[i] = o(x)
        return self._scalarize(fs)

    def lower_bound(self) -> float | int:
        """
        Get the lower bound of the scalarization result.

        This function returns a theoretical limit for how good a solution
        could be at best. If no real limit is known, the function returns
        `-inf`.

        :return: the lower bound of the scalarization result
        """
        return self.__lower_bound

    def upper_bound(self) -> float | int:
        """
        Get the upper bound of the scalarization result.

        This function returns a theoretical limit for how bad a solution could
        be at worst. If no real limit is known, the function returns `inf`.

        :return: the upper bound of the scalarization result
        """
        return self.__upper_bound

    def evaluate(self, x) -> float | int:
        """
        Convert the multi-objective problem into a single-objective one.

        This function first evaluates all encapsulated objectives and then
        scalarizes the result.

        :param x: the candidate solution
        :returns: the scalarized objective value
        """
        return self.f_evaluate(x, self._temp)

    def __str__(self) -> str:
        """Get the string representation of this basic scalarization."""
        return "basicMoProblem"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this function to the provided destination.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_SPACE_NUM_VARS, self.__dimension)
        logger.key_value(KEY_NUMPY_TYPE, numpy_type_to_str(self.__dtype))
        for i, o in enumerate(self._objectives):
            with logger.scope(f"{SCOPE_OBJECTIVE_FUNCTION}{i}") as scope:
                o.log_parameters_to(scope)

    def validate(self, x: np.ndarray) -> None:
        """
        Validate an objective vector.

        :param x: the objective vector
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().f_validate(x)

        lb: Final[tuple[int | float, ...]] = self.__lower_bounds
        ub: Final[tuple[int | float, ...]] = self.__upper_bounds
        for i, v in enumerate(x):
            if v < lb[i]:
                raise ValueError(
                    f"encountered {v} at index {i} of {x}, which is below the "
                    f"lower bound {lb[i]} for that position.")
            if v > ub[i]:
                raise ValueError(
                    f"encountered {v} at index {i} of {x}, which is above the "
                    f"upper bound {ub[i]} for that position.")
